package gov.usgs.earthquake.nshmp.www;

import java.nio.file.Path;
import java.util.List;

import gov.usgs.earthquake.nshmp.www.fault.NshmFaultSection;
import gov.usgs.earthquake.nshmp.www.services.FaultSectionsService;
import gov.usgs.earthquake.nshmp.www.services.FaultSectionsService.Query;

import io.micronaut.context.annotation.Value;
import io.micronaut.core.annotation.Nullable;
import io.micronaut.http.HttpRequest;
import io.micronaut.http.HttpResponse;
import io.micronaut.http.MediaType;
import io.micronaut.http.annotation.Body;
import io.micronaut.http.annotation.Controller;
import io.micronaut.http.annotation.Get;
import io.micronaut.http.annotation.PathVariable;
import io.micronaut.http.annotation.Post;
import io.micronaut.http.annotation.QueryValue;
import io.swagger.v3.oas.annotations.Operation;
import io.swagger.v3.oas.annotations.media.Content;
import io.swagger.v3.oas.annotations.media.Schema;
import io.swagger.v3.oas.annotations.responses.ApiResponse;
import io.swagger.v3.oas.annotations.tags.Tag;
import jakarta.inject.Inject;

@Tag(name = "Fault Sections",
    description = "Fault section services")
@Controller("/fault-sections")
public class FaultSectionsController {

  @Inject
  private NshmpMicronautServlet servlet;

  @Value("${faults-path:" + Utils.FAULT_RESOURCES + "}")
  Path path;

  /**
   * GET method that returns the fault section based on group, id, or name.
   *
   * @param request The HTTP request
   * @param group The fault section group
   * @param id The fault section id
   * @param name The fault section name
   * @param raw Whether to return raw GeoJSON feature collection
   */
  @Operation(
      summary = "Returns the fault section based on a group, name, or id.",
      description = "Returns a fault section based on a group, name, **OR** id.\n\n" +
          "For supported groups and names see the usage information.\n\n" +
          "Given no query parameters the usage information is returned.",
      operationId = "fault-sections_doGetFaults")
  @ApiResponse(
      description = "Fault section",
      responseCode = "200",
      content = @Content(
          schema = @Schema(type = "string")))
  @Get(uri = "{?group,id,name,faults,raw}", produces = MediaType.APPLICATION_JSON)
  public HttpResponse<String> doGetFaults(
      HttpRequest<?> request,
      @QueryValue @Nullable List<String> group,
      @QueryValue @Nullable List<Integer> id,
      @QueryValue @Nullable List<String> name,
      @QueryValue @Nullable NshmFaultSection faults,
      @QueryValue @Nullable Boolean raw) {
    return FaultSectionsService.handleDoGetFaults(request, group, id, name, faults, raw, path);
  }

  /**
   * Slash delimited GET method that returns the fault section based on group or
   * id.
   *
   * @param request The HTTP request
   * @param value The group or id of the fault section
   * @param raw Whether to return raw GeoJSON feature collection
   */
  @Operation(
      summary = "Returns the fault section based on a group or id.",
      description = "Returns a fault section.\n\n" +
          "For supported groups and names see the usage information.\n\n" +
          "Given no query parameters the usage information is returned.",
      operationId = "fault-sections_doGetFaultsSlash")
  @ApiResponse(
      description = "Fault section",
      responseCode = "200",
      content = @Content(
          schema = @Schema(type = "string")))
  @Get(uri = "/{value}{/faults}{/raw}", produces = MediaType.APPLICATION_JSON)
  public HttpResponse<String> doGetFaultsSlash(
      HttpRequest<?> request,
      @PathVariable @Nullable String value,
      @PathVariable @Nullable NshmFaultSection faults,
      @PathVariable @Nullable Boolean raw) {
    return FaultSectionsService.handleDoGetFaultsSlash(request, value, faults, raw, path);
  }

  /**
   * POST method that returns the fault section.
   *
   * @param request The HTTO request
   * @param query The query
   */
  @Operation(
      summary = "Returns the fault section based on a group or id.",
      description = "Returns a fault section.\n\n" +
          "For supported groups and names see the usage information.\n\n" +
          "Given no query parameters the usage information is returned.",
      operationId = "fault-sections_doPost")
  @ApiResponse(
      description = "Fault section",
      responseCode = "200",
      content = @Content(
          schema = @Schema(type = "string")))
  @Post(consumes = MediaType.APPLICATION_JSON, produces = MediaType.APPLICATION_JSON)
  public HttpResponse<String> doPost(HttpRequest<?> request, @Body @Nullable Query query) {
    return FaultSectionsService.handleDoPost(request, query, path);
  }

}
