package gov.usgs.earthquake.nshmp.www;

import gov.usgs.earthquake.nshmp.www.gps.GpsDataSet;
import gov.usgs.earthquake.nshmp.www.services.GpsService;

import io.micronaut.core.annotation.Nullable;
import io.micronaut.http.HttpRequest;
import io.micronaut.http.HttpResponse;
import io.micronaut.http.MediaType;
import io.micronaut.http.annotation.Controller;
import io.micronaut.http.annotation.Get;
import io.micronaut.http.annotation.PathVariable;
import io.micronaut.http.annotation.QueryValue;
import io.swagger.v3.oas.annotations.Operation;
import io.swagger.v3.oas.annotations.media.Content;
import io.swagger.v3.oas.annotations.media.Schema;
import io.swagger.v3.oas.annotations.responses.ApiResponse;
import io.swagger.v3.oas.annotations.tags.Tag;
import jakarta.inject.Inject;

/**
 * GPS data service to return GPS data for NSHM.
 *
 * <p> Model GPS data can be returned as CSV or GeoJSON strings.
 *
 * @author U.S. Geological Survey
 */
@Tag(name = "GPS Services")
@Controller("/gps")
public class GpsController {

  @Inject
  private NshmpMicronautServlet servlet;

  /**
   * GET method that returns the GPS results for the specified model.
   *
   * @param request The HTTP request
   * @param model The NSHM model
   * @param format The requested data format
   */
  @Operation(
      summary = "Returns usage information or " +
          "GPS dataset for the specified NSHM model",
      description = "Returns a NSHM GPS dataset with velocities in mm/yr and, if applicable, " +
          "vertical components are positive UP.\n\n" +
          "For supported datasets and formats see the usage information\n\n" +
          "Given no query parameters the usage information is returned",
      operationId = "gps_doGetGps")
  @ApiResponse(
      description = "NSHM GPS datasets",
      responseCode = "200",
      content = @Content(
          schema = @Schema(type = "string")))
  @Get(uri = "{?model,format}", produces = MediaType.APPLICATION_JSON)
  public HttpResponse<String> doGetGpsData(
      HttpRequest<?> request,
      @QueryValue @Nullable GpsDataSet model,
      @QueryValue @Nullable GpsService.GpsFormat format) {
    return GpsService.handleDoGetGps(request, model, format);
  }

  /**
   * Slash delimited get method.
   *
   * @param request The HTTP request
   * @param model The GPS dataset
   * @param format Data format
   */
  @Operation(
      summary = "Returns all GPS data for a specific dataset",
      description = "Returns a NSHM GPS dataset with velocities in mm/yr and, if applicable, " +
          "vertical components are positive UP.\n\n" +
          "For supported datasets and formats see the usage information\n\n",
      operationId = "gps_doGetGpsSlash")
  @ApiResponse(
      description = "NSHM GPS datasets",
      responseCode = "200",
      content = @Content(
          schema = @Schema(type = "string")))
  @Get(uri = "/{model}{/format}", produces = MediaType.APPLICATION_JSON)
  public HttpResponse<String> doGetGpsDataSlash(
      HttpRequest<?> request,
      @PathVariable @Nullable GpsDataSet model,
      @PathVariable @Nullable GpsService.GpsFormat format) {
    return doGetGpsData(request, model, format);
  }

}
