package gov.usgs.earthquake.nshmp.www;

import gov.usgs.earthquake.nshmp.www.services.GulfService;

import io.micronaut.core.annotation.Nullable;
import io.micronaut.http.HttpRequest;
import io.micronaut.http.HttpResponse;
import io.micronaut.http.MediaType;
import io.micronaut.http.annotation.Controller;
import io.micronaut.http.annotation.Get;
import io.micronaut.http.annotation.PathVariable;
import io.micronaut.http.annotation.QueryValue;
import io.swagger.v3.oas.annotations.Operation;
import io.swagger.v3.oas.annotations.media.Content;
import io.swagger.v3.oas.annotations.media.Schema;
import io.swagger.v3.oas.annotations.responses.ApiResponse;
import io.swagger.v3.oas.annotations.tags.Tag;
import jakarta.inject.Inject;

/**
 * Gulf sediment depth data service.
 *
 * <p> Note: If the longitude and latitude supplied in the query is not
 * contained in the dataset region, the resulting depth value is null.
 *
 * <p> Note: The supplied longitude and latitude values are rounded to the
 * nearest grid point ({@code 0.05} degrees).
 *
 * @author U.S. Geological Survey
 */
@Tag(name = "Gulf Services")
@Controller("/gulf")
public class GulfController {

  @Inject
  private NshmpMicronautServlet servlet;

  /**
   * GET method that returns the gulf sediment depth results based on longitude
   * and latitude. Returns usage if query is empty.
   *
   * @param request The HTTP request
   * @param longitude The longitude of interest
   * @param latitude The latitude of interest
   */
  @Operation(
      summary = "Returns usage information or " +
          "sediment depth (m) based on a longitude and latitude.",
      description = "Returns sediment depth (m) for a given site (longitude, latitude).\n\n" +
          "For supported dataset region see the usage information.\n\n" +
          "Given no query parameters the usage information is returned.",
      operationId = "gulf_doGetGulf")
  @ApiResponse(
      description = "Gulf coast sediment depth",
      responseCode = "200",
      content = @Content(
          schema = @Schema(type = "string")))
  @Get(uri = "{?longitude,latitude}", produces = MediaType.APPLICATION_JSON)
  public HttpResponse<String> doGetGulf(
      HttpRequest<?> request,
      @QueryValue @Nullable Double longitude,
      @QueryValue @Nullable Double latitude) {
    return GulfService.handleDoGetGulf(request, longitude, latitude);

  }

  /**
   * Slash delimited get method.
   *
   * @param request The HTTP request
   * @param longitude The longitude in degrees
   * @param latitude The latitude in degrees
   */
  @Operation(
      summary = "Returns sediment depth (m) based on a longitude and latitude",
      description = "Returns the sediment depth (m) for specified longitude and latitude.\n\n",
      operationId = "gulf_doGetGulfSlash")
  @ApiResponse(
      description = "Gulf coast sediment depth",
      responseCode = "200",
      content = @Content(
          schema = @Schema(type = "string")))
  @Get(uri = "/{longitude}/{latitude}", produces = MediaType.APPLICATION_JSON)
  public HttpResponse<String> doGetGulfSlash(HttpRequest<?> request,
      @PathVariable @Nullable Double longitude,
      @PathVariable @Nullable Double latitude) {
    return doGetGulf(request, longitude, latitude);
  }

}
