package gov.usgs.earthquake.nshmp.www;

import static com.google.common.base.CaseFormat.LOWER_CAMEL;
import static com.google.common.base.CaseFormat.UPPER_UNDERSCORE;

import java.lang.reflect.Type;

import com.google.common.base.Enums;
import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import com.google.gson.JsonArray;
import com.google.gson.JsonElement;
import com.google.gson.JsonObject;
import com.google.gson.JsonSerializationContext;
import com.google.gson.JsonSerializer;

import gov.usgs.earthquake.nshmp.internal.UsRegion;
import gov.usgs.earthquake.nshmp.www.fault.FaultGroup;
import gov.usgs.earthquake.nshmp.www.meta.StringParameter;

import io.micronaut.http.HttpResponse;

public class Utils {

  public static final String FAULT_RESOURCES = "./libs";
  public static final String GPS_RESOURCES = "/gps/";
  public static final String GULF_RESOURCES = "/gulf/";
  public static final Gson GSON;

  static {
    GSON = new GsonBuilder()
        .registerTypeAdapter(StringParameter.class, new StringParameterSerializer())
        .disableHtmlEscaping()
        .serializeNulls()
        .setPrettyPrinting()
        .create();
  }

  public enum Key {
    DISTANCE,
    ID,
    IMT,
    GMM,
    GPSDATASET,
    GROUP,
    LABEL,
    LATITUDE,
    LONGITUDE,
    M_MIN,
    M_MAX,
    MODEL,
    NAME,
    R_MAX,
    R_MIN,
    STEP,
    Z1P0,
    Z2P5;

    @Override
    public String toString() {
      return UPPER_UNDERSCORE.to(LOWER_CAMEL, name());
    }
  }

  public static HttpResponse<String> handleError(
      Throwable e,
      String name,
      String url) {
    var msg = e.getMessage() + " (see logs)";
    var svcResponse = ResponseBody.error()
        .name(name)
        .url(url)
        .metadata(new ResponseMetadata(WsVersion.appVersions()))
        .request(msg)
        .response(url)
        .build();
    var response = GSON.toJson(svcResponse);
    e.printStackTrace();
    return HttpResponse.serverError(response);
  }

  private static final class StringParameterSerializer implements JsonSerializer<StringParameter> {
    @Override
    public JsonElement serialize(
        StringParameter stringParameter,
        Type typeOfSrc,
        JsonSerializationContext context) {
      JsonObject json = new JsonObject();
      json.addProperty(Key.LABEL.toString(), stringParameter.label);

      JsonArray jsonValues = new JsonArray();

      for (String value : stringParameter.values) {
        JsonObject jsonValuesObj = new JsonObject();
        var valueUC = value.toUpperCase();

        if (Enums.getIfPresent(UsRegion.class, valueUC).isPresent()) {
          UsRegion region = UsRegion.valueOf(valueUC);
          jsonValuesObj.addProperty(Key.LABEL.toString(), region.toString());
          jsonValuesObj.addProperty(Key.ID.toString(), region.name());
          jsonValues.add(jsonValuesObj);
        } else if (Enums.getIfPresent(FaultGroup.class, valueUC).isPresent()) {
          FaultGroup fault = FaultGroup.valueOf(valueUC);
          jsonValuesObj.addProperty(Key.LABEL.toString(), fault.toString());
          jsonValuesObj.addProperty(Key.ID.toString(), fault.name());
          jsonValues.add(jsonValuesObj);
        } else {
          jsonValues.add(value);
        }
      }

      json.add("values", jsonValues);

      return json;
    }
  }
}
