package gov.usgs.earthquake.nshmp.www.fault;

import static com.google.common.base.Preconditions.checkState;

import java.io.IOException;
import java.net.MalformedURLException;
import java.nio.file.FileVisitResult;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.SimpleFileVisitor;
import java.nio.file.attribute.BasicFileAttributes;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import com.google.common.collect.Multimap;
import com.google.common.collect.MultimapBuilder;

import gov.usgs.earthquake.nshmp.internal.UsRegion;
import gov.usgs.earthquake.nshmp.model.SourceFeature;

public class FaultSections {

  private static String GEOJSON_SUFFIX = ".geojson";
  private Map<Integer, SourceFeature> faultsById = new HashMap<>();
  private Map<String, SourceFeature> faultsByName = new HashMap<>();
  private Multimap<UsRegion, SourceFeature> faultsByRegion;

  private FaultSections(
      Map<Integer, SourceFeature> ids,
      Map<String, SourceFeature> names,
      Multimap<UsRegion, SourceFeature> regions) {
    faultsById = ids;
    faultsByName = names;
    faultsByRegion = regions;
  }

  /**
   * Load all the fault sections from a given path.
   *
   * @param path The path to where all fault sections directories are
   */
  public static Map<NshmFaultSection, FaultSections> getFaultSections(Path path) {
    var faultSectionsMap = new HashMap<NshmFaultSection, FaultSections>();

    for (var nshmFaultSection : NshmFaultSection.values()) {
      faultSectionsMap.put(
          nshmFaultSection,
          loadFaultSections(path.resolve(nshmFaultSection.directory())));
    }

    return Map.copyOf(faultSectionsMap);
  }

  /**
   * Returns {@code NshmFaultSection} by id
   * @param ids The ids to get
   */
  public List<SourceFeature> ids(List<Integer> ids) {
    return ids.stream()
        .map(id -> {
          if (!faultsById.containsKey(id)) {
            throw new IllegalArgumentException("Id [" + id + "] not found");
          }
          return faultsById.get(id);
        })
        .collect(Collectors.toList());
  }

  /**
   * Returns {@code NshmFaultSection} by name
   * @param name The fault name
   */
  public List<SourceFeature> names(List<String> names) {
    return names.stream()
        .map(name -> {
          if (!faultsByName.containsKey(name)) {
            throw new IllegalArgumentException("Name [" + name + "] not found");
          }
          return faultsByName.get(name);
        })
        .sorted((s1, s2) -> s1.name.compareTo(s2.name))
        .collect(Collectors.toList());
  }

  /**
   * Returns {@code NshmFaultSection} by region
   * @param regions The regions to get
   */
  public List<SourceFeature> regions(UsRegion... regions) {
    return Stream.of(regions)
        .flatMap(region -> {
          if (!faultsByRegion.containsKey(region)) {
            throw new IllegalArgumentException("Region [" + region + "] not found");
          }
          return faultsByRegion.get(region).stream();
        })
        .sorted((s1, s2) -> s1.state.orElseThrow().compareTo(s2.state.orElseThrow()))
        .collect(Collectors.toList());
  }

  /**
   * Return the set of ids in the fault sections
   */
  public Set<Integer> ids() {
    return new TreeSet<>(faultsById.keySet());
  }

  /**
   * Return the set of fault names in the fault sections
   */
  public Set<String> names() {
    return new TreeSet<>(faultsByName.keySet());
  }

  /**
   * Return the set of {@code UsRegion} in the fault sections
   */
  public Set<UsRegion> regions() {
    return new TreeSet<>(faultsByRegion.keySet());
  }

  private static FaultSections loadFaultSections(Path path) {
    path = path.toAbsolutePath().normalize();

    if (!Files.exists(path)) {
      throw new IllegalArgumentException(
          "Fault sections path [" + path + "] does not exist");
    }

    try {
      FaultFinder finder = new FaultFinder();
      Files.walkFileTree(path, finder);
      return new FaultSections(finder.ids, finder.names, finder.regions);
    } catch (IOException e) {
      throw new RuntimeException(e);
    }
  }

  private static class FaultFinder extends SimpleFileVisitor<Path> {
    Map<Integer, SourceFeature> ids;
    Map<String, SourceFeature> names;
    Multimap<UsRegion, SourceFeature> regions;

    FaultFinder() {
      ids = new HashMap<>();
      names = new HashMap<>();
      regions = MultimapBuilder
          .enumKeys(UsRegion.class)
          .arrayListValues()
          .build();
    }

    @Override
    public FileVisitResult visitFile(Path path, BasicFileAttributes attrs)
        throws MalformedURLException {

      if (path.toString().contains(GEOJSON_SUFFIX)) {

        boolean isInterface = path.toString().contains("(interface)");
        boolean isDecollement = path.toString().contains("Decollement");

        SourceFeature feature = isInterface
            ? SourceFeature.newInterfaceSection(path)
            : isDecollement
                ? SourceFeature.newDecollementSection(path)
                : SourceFeature.newFaultSection(path);

        int id = feature.id;
        checkState(
            !ids.containsKey(id),
            "Duplicate id: [%s] %s : %s ",
            id, feature.state.orElseThrow(), feature.name);
        ids.put(id, feature);

        String name = String.format("%s - %s", feature.name, feature.state.orElseThrow());
        names.put(name, feature);

        UsRegion region = UsRegion.valueOf(feature.state.orElseThrow());
        regions.put(region, ids.get(id));

        if (feature.states.size() > 1) {
          UsRegion secondaryRegion = UsRegion.valueOf(feature.states.get(1));
          regions.put(secondaryRegion, feature);
        }
      }
      return FileVisitResult.CONTINUE;
    }
  }

}
