package gov.usgs.earthquake.nshmp.www.gmm;

import static java.lang.Math.cos;
import static java.lang.Math.hypot;
import static java.lang.Math.sin;
import static java.lang.Math.tan;
import static java.lang.Math.toRadians;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.EnumMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import com.google.common.collect.ImmutableList;
import com.google.common.primitives.Doubles;

import gov.usgs.earthquake.nshmp.Maths;
import gov.usgs.earthquake.nshmp.gmm.Gmm;
import gov.usgs.earthquake.nshmp.gmm.GmmInput;
import gov.usgs.earthquake.nshmp.gmm.GroundMotion;
import gov.usgs.earthquake.nshmp.gmm.GroundMotionModel;
import gov.usgs.earthquake.nshmp.gmm.GroundMotions;
import gov.usgs.earthquake.nshmp.gmm.Imt;
import gov.usgs.earthquake.nshmp.tree.Branch;
import gov.usgs.earthquake.nshmp.tree.LogicTree;
import gov.usgs.earthquake.nshmp.www.gmm.GmmService.Distance;
import gov.usgs.earthquake.nshmp.www.gmm.GmmService.Magnitude;
import gov.usgs.earthquake.nshmp.www.gmm.GmmService.Request;

import jakarta.inject.Singleton;

/*
 * GMM service calculators.
 *
 * @author U.S. Geological Survey
 */
@Singleton
class GmmCalc {

  private static final double PGA_PERIOD = 0.001;

  /* Compute ground motion response spectra. */
  static Map<Gmm, GmmData> spectra(Request request, boolean commonImts) {

    /*
     * NOTE: At present, program assumes that all supplied Gmms support PGA.
     * Although most currently implemented models do, this may not be the case
     * in the future and program may produce unexpected results.
     */

    /* Common imts and periods; may not be used. */
    Set<Imt> saImts = Gmm.responseSpectrumImts(request.gmms);
    List<Double> periods = new ArrayList<>();
    periods.add(PGA_PERIOD);
    periods.addAll(Imt.periods(saImts));

    Map<Gmm, GmmData> gmmSpectra = new EnumMap<>(Gmm.class);
    for (Gmm gmm : request.gmms) {
      if (!commonImts) {
        saImts = gmm.responseSpectrumImts();
        periods = ImmutableList.<Double> builder()
            .add(PGA_PERIOD)
            .addAll(Imt.periods(saImts))
            .build();
      }

      List<LogicTree<GroundMotion>> imtTrees = new ArrayList<>();
      imtTrees.add(gmm.instance(Imt.PGA).calc(request.input));
      for (Imt imt : saImts) {
        imtTrees.add(gmm.instance(imt).calc(request.input));
      }

      GmmData dataGroup = treesToDataGroup(periods, imtTrees);
      gmmSpectra.put(gmm, dataGroup);
    }
    return gmmSpectra;
  }

  private static GmmData treesToDataGroup(
      List<Double> xValues,
      List<LogicTree<GroundMotion>> trees) {

    // Can't use Trees.transpose() because some
    // GMMs have period dependent weights

    LogicTree<GroundMotion> modelTree = trees.get(0);
    List<String> branchIds = modelTree.stream()
        .map(Branch::id)
        .collect(Collectors.toList());

    double[] xs = xValues.stream().mapToDouble(Double::valueOf).toArray();
    double[] μs = new double[xValues.size()];
    double[] σs = new double[xValues.size()];

    // build combined μ and σ
    for (int i = 0; i < trees.size(); i++) {
      GroundMotion combined = GroundMotions.combine(trees.get(i));
      μs[i] = combined.mean();
      σs[i] = combined.sigma();
    }

    List<EpiBranch> epiBranches = new ArrayList<>();

    // short circuit if tree is single branch
    if (modelTree.size() > 1) {

      // branch index, imt index
      List<double[]> μBranches = new ArrayList<>();
      List<double[]> σBranches = new ArrayList<>();
      List<double[]> weights = new ArrayList<>();
      for (int i = 0; i < modelTree.size(); i++) {
        μBranches.add(new double[xValues.size()]);
        σBranches.add(new double[xValues.size()]);
        weights.add(new double[xValues.size()]);
      }

      // imt indexing
      for (int i = 0; i < trees.size(); i++) {
        LogicTree<GroundMotion> tree = trees.get(i);
        // epi branch indexing
        for (int j = 0; j < tree.size(); j++) {
          Branch<GroundMotion> branch = tree.get(j);
          μBranches.get(j)[i] = branch.value().mean();
          σBranches.get(j)[i] = branch.value().sigma();
          weights.get(j)[i] = branch.weight();
        }
      }

      for (int i = 0; i < modelTree.size(); i++) {
        EpiBranch epiBranch = new EpiBranch(
            branchIds.get(i),
            μBranches.get(i),
            σBranches.get(i),
            weights.get(i));
        epiBranches.add(epiBranch);
      }
    }

    GmmData gmmData = new GmmData(xs, μs, σs, epiBranches);
    return gmmData;
  }

  /* Compute ground motions over a range of distances. */
  static Map<Gmm, GmmData> distance(Distance.Request request, double[] distances) {
    var inputList = hangingWallDistances(request.input, distances);
    return calculateGroundMotions(request.gmms, inputList, request.imt, distances);
  }

  /* Compute ground motions over a range of magnitudes. */
  static Map<Gmm, GmmData> magnitude(
      Magnitude.Request request,
      double[] magnitudes,
      double distance) {

    var gmmInputs = Arrays.stream(magnitudes)
        .mapToObj(Mw -> GmmInput.builder().fromCopy(request.input).mag(Mw).build())
        .map(gmmInput -> hangingWallDistance(gmmInput, distance))
        .collect(Collectors.toList());
    return calculateGroundMotions(request.gmms, gmmInputs, request.imt, magnitudes);
  }

  private static List<GmmInput> hangingWallDistances(
      GmmInput inputModel,
      double[] rValues) {

    return Arrays.stream(rValues)
        .mapToObj(r -> hangingWallDistance(inputModel, r))
        .collect(Collectors.toList());
  }

  /* Compute distance metrics for a fault. */
  private static GmmInput hangingWallDistance(GmmInput in, double r) {
    /* Dip in radians */
    double δ = toRadians(in.dip);

    /* Horizontal and vertical widths of fault */
    double h = cos(δ) * in.width;
    double v = sin(δ) * in.width;

    /* Depth to bottom of rupture */
    double zBot = in.zTor + v;

    /* Distance range over which site is normal to fault plane */
    double rCutLo = tan(δ) * in.zTor;
    double rCutHi = tan(δ) * zBot + h;

    /* rRup values corresponding to cutoffs above */
    double rRupLo = Maths.hypot(in.zTor, rCutLo);
    double rRupHi = Maths.hypot(zBot, rCutHi - h);

    double rJB = (r < 0) ? -r : (r < h) ? 0.0 : r - h;
    double rRup = (r < rCutLo)
        ? hypot(r, in.zTor)
        : (r > rCutHi)
            ? hypot(r - h, zBot)
            : rRupScaled(
                r, rCutLo, rCutHi, rRupLo, rRupHi);

    return GmmInput.builder()
        .fromCopy(in)
        .distances(rJB, rRup, r)
        .build();
  }

  private static Map<Gmm, GmmData> calculateGroundMotions(
      Set<Gmm> gmms,
      List<GmmInput> gmmInputs,
      Imt imt,
      double[] distances) {

    Map<Gmm, GmmData> gmValues = new EnumMap<>(Gmm.class);

    for (Gmm gmm : gmms) {
      List<LogicTree<GroundMotion>> trees = new ArrayList<>();
      GroundMotionModel model = gmm.instance(imt);
      for (GmmInput gmmInput : gmmInputs) {
        trees.add(model.calc(gmmInput));
      }
      GmmData dataGroup = GmmCalc.treesToDataGroup(Doubles.asList(distances), trees);
      gmValues.put(gmm, dataGroup);
    }
    return gmValues;
  }

  /*
   * Computes rRup for a surface distance r. The range [rCutLo, rCutHi] must
   * contain r; rRupLo and rRupHi are rRup at rCutLo and rCutHi, respectively.
   */
  private static double rRupScaled(
      double r,
      double rCutLo,
      double rCutHi,
      double rRupLo,
      double rRupHi) {

    double rRupΔ = rRupHi - rRupLo;
    double rCutΔ = rCutHi - rCutLo;
    return rRupLo + (r - rCutLo) / rCutΔ * rRupΔ;
  }

  static class GmmData {

    final double[] xs;
    final double[] μs;
    final double[] σs;
    final List<EpiBranch> tree;

    GmmData(
        double[] xs,
        double[] μs,
        double[] σs,
        List<EpiBranch> tree) {

      this.xs = xs;
      this.μs = μs;
      this.σs = σs;
      this.tree = tree;
    }
  }

  static class EpiBranch {

    final String id;
    final double[] μs;
    final double[] σs;
    final double[] weights;

    EpiBranch(String id, double[] μs, double[] σs, double[] weights) {
      this.id = id;
      this.μs = μs;
      this.σs = σs;
      this.weights = weights;
    }
  }

}
