package gov.usgs.earthquake.nshmp.www.gmm;

import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.DIP;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.MW;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.RAKE;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.RJB;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.RRUP;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.RX;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.VS30;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.VSINF;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.WIDTH;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.Z1P0;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.Z2P5;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.ZHYP;
import static gov.usgs.earthquake.nshmp.gmm.GmmInput.Field.ZTOR;
import static gov.usgs.earthquake.nshmp.gmm.Imt.AI;
import static gov.usgs.earthquake.nshmp.gmm.Imt.PGV;
import static io.micronaut.core.type.Argument.BOOLEAN;
import static io.micronaut.core.type.Argument.DOUBLE;
import static java.util.stream.Collectors.toCollection;

import java.lang.reflect.Type;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.EnumSet;
import java.util.List;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;

import com.google.common.collect.Range;
import com.google.common.primitives.Doubles;
import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import com.google.gson.JsonElement;
import com.google.gson.JsonObject;
import com.google.gson.JsonSerializationContext;
import com.google.gson.JsonSerializer;

import gov.usgs.earthquake.nshmp.data.DoubleData;
import gov.usgs.earthquake.nshmp.data.Sequences;
import gov.usgs.earthquake.nshmp.gmm.Gmm;
import gov.usgs.earthquake.nshmp.gmm.GmmInput;
import gov.usgs.earthquake.nshmp.gmm.GmmInput.Constraints;
import gov.usgs.earthquake.nshmp.gmm.GmmInput.Field;
import gov.usgs.earthquake.nshmp.gmm.Imt;
import gov.usgs.earthquake.nshmp.www.ResponseBody;
import gov.usgs.earthquake.nshmp.www.ResponseMetadata;
import gov.usgs.earthquake.nshmp.www.WsUtils;
import gov.usgs.earthquake.nshmp.www.WsVersion;
import gov.usgs.earthquake.nshmp.www.gmm.GmmService.Id;
import gov.usgs.earthquake.nshmp.www.meta.EnumParameter;

import io.micronaut.http.HttpParameters;
import io.micronaut.http.HttpRequest;
import io.micronaut.http.HttpResponse;

class ServiceUtil {

  final static int ROUND = 5;
  static final Range<Double> MAGNITUDE_DEFAULT_VALUES = Range.closed(5.0, 8.0);
  static final double MAGNITUDE_DEFAULT_STEP = 0.1;
  static final double MAGNITUDE_DEFAULT_DISTANCE = 10.0;
  static final Range<Double> DISTANCE_DEFAULT_VALUES = Range.closed(-100.0, 100.0);

  static final Gson GSON;

  static {
    GSON = new GsonBuilder()
        .setPrettyPrinting()
        .serializeNulls()
        .disableHtmlEscaping()
        .registerTypeAdapter(Double.class, new WsUtils.NaNSerializer())
        .registerTypeAdapter(ServiceUtil.Parameters.class,
            new ServiceUtil.Parameters.Serializer())
        .registerTypeAdapter(Imt.class, new WsUtils.EnumSerializer<Imt>())
        .registerTypeAdapter(Constraints.class, new WsUtils.ConstraintsSerializer())
        .create();
  }

  public static HttpResponse<String> metadata(HttpRequest<?> request, Id service) {
    var metadata = GSON.toJson(ServiceUtil.getMetadata(request, service));
    return HttpResponse.ok(metadata);
  }

  /** Query and JSON reqest/response keys. */
  static final class Key {
    public static final String DISTANCE = "distance";
    public static final String IMT = "imt";
    public static final String GMM = "gmm";
    public static final String M_MIN = "mMin";
    public static final String M_MAX = "mMax";
    public static final String R_MAX = "rMax";
    public static final String R_MIN = "rMin";
    public static final String STEP = "step";
  }

  /* Read the GMM input query values. */
  static GmmInput readGmmInput(HttpRequest<?> request) {
    HttpParameters params = request.getParameters();
    GmmInput.Builder b = GmmInput.builder().withDefaults();
    params.getFirst(MW.id, DOUBLE).ifPresent(b::mag);
    params.getFirst(RJB.id, DOUBLE).ifPresent(b::rJB);
    params.getFirst(RRUP.id, DOUBLE).ifPresent(b::rRup);
    params.getFirst(RX.id, DOUBLE).ifPresent(b::rX);
    params.getFirst(DIP.id, DOUBLE).ifPresent(b::dip);
    params.getFirst(WIDTH.id, DOUBLE).ifPresent(b::width);
    params.getFirst(ZTOR.id, DOUBLE).ifPresent(b::zTor);
    params.getFirst(ZHYP.id, DOUBLE).ifPresent(b::zHyp);
    params.getFirst(RAKE.id, DOUBLE).ifPresent(b::rake);
    params.getFirst(VS30.id, DOUBLE).ifPresent(b::vs30);
    params.getFirst(VSINF.id, BOOLEAN).ifPresent(b::vsInf);
    params.getFirst(Z1P0.id, DOUBLE).ifPresent(b::z1p0);
    params.getFirst(Z2P5.id, DOUBLE).ifPresent(b::z2p5);
    return b.build();
  }

  /* Read the 'gmm' query values. */
  static Set<Gmm> readGmms(HttpRequest<?> request) {
    return request.getParameters()
        .getAll(Key.GMM)
        .stream()
        .map(s -> s.split(","))
        .flatMap(Arrays::stream)
        .map(Gmm::valueOf)
        .collect(toCollection(() -> EnumSet.noneOf(Gmm.class)));
  }

  static ResponseBody<String, MetadataResponse> getMetadata(
      HttpRequest<?> request,
      Id service) {
    String url = request.getUri().getPath();
    return ResponseBody.<String, MetadataResponse> usage()
        .name(service.name)
        .url(url)
        .metadata(new ResponseMetadata(WsVersion.appVersions()))
        .request(url)
        .response(new MetadataResponse(request, service))
        .build();
  }

  static double[] sequenceLog(double min, double max, double step) {
    return DoubleData.round(
        ROUND,
        DoubleData.pow10(
            Sequences.arrayBuilder(Math.log10(min), Math.log10(max), step)
                .centered()
                .build()));
  }

  static double[] sequenceLinear(double min, double max, double step) {
    return Sequences.arrayBuilder(min, max, step)
        .scale(ROUND)
        .centered()
        .build();
  }

  static final class MetadataResponse {
    String description;
    String syntax;
    Parameters parameters;

    public MetadataResponse(HttpRequest<?> request, Id service) {
      this.syntax = request.getUri().getPath();
      this.description = service.description;
      this.parameters = new Parameters(service);
    }
  }

  @SuppressWarnings("unchecked")
  private static Param createGmmInputParam(
      Field field,
      Optional<?> constraint) {

    return (field == VSINF)
        ? new BooleanParam(field)
        : new NumberParam(field, (Range<Double>) constraint.orElseThrow());
  }

  /*
   * Placeholder class; all parameter serialization is done via the custom
   * Serializer. Service reference needed serialize().
   */
  static final class Parameters {

    private final Id service;

    Parameters(Id service) {
      this.service = service;
    }

    static final class Serializer implements JsonSerializer<Parameters> {

      @Override
      public JsonElement serialize(
          Parameters meta,
          Type type,
          JsonSerializationContext context) {

        JsonObject root = new JsonObject();

        if (!meta.service.equals(Id.SPECTRA)) {
          Set<Imt> imtSet = EnumSet.complementOf(EnumSet.range(PGV, AI));
          final EnumParameter<Imt> imts = new EnumParameter<>(
              "Intensity measure type",
              imtSet);
          root.add(Key.IMT, context.serialize(imts));
        }

        /* Serialize input fields. */
        Constraints defaults = Constraints.defaults();
        for (Field field : Field.values()) {
          if (!meta.service.equals(Id.SPECTRA) &&
              (field.equals(Field.RX) || field.equals(Field.RRUP) || field.equals(Field.RJB))) {
            continue;
          }
          Param param = createGmmInputParam(field, defaults.get(field));
          JsonElement fieldElem = context.serialize(param);
          root.add(field.id, fieldElem);
        }

        if (meta.service.equals(Id.DISTANCE) || meta.service.equals(Id.HW_FW)) {
          double min = meta.service.equals(Id.HW_FW)
              ? Range.openClosed(0.0, 1.0).lowerEndpoint()
              : DISTANCE_DEFAULT_VALUES.lowerEndpoint();
          NumberParam rMin = new NumberParam(
              "Minimum distance",
              "The minimum distance to use for calculation",
              Field.RX.units.orElse(null),
              Range.openClosed(0.0, 1000.0),
              min);
          root.add(Key.R_MIN, context.serialize(rMin));

          NumberParam rMax = new NumberParam(
              "Maximum distance",
              "The maximum distance to use for calculation",
              Field.RX.units.orElse(null),
              Range.closed(-1000.0, 1000.0),
              DISTANCE_DEFAULT_VALUES.upperEndpoint());
          root.add(Key.R_MAX, context.serialize(rMax));
        }

        if (meta.service.equals(Id.MAGNITUDE)) {
          @SuppressWarnings("unchecked")
          NumberParam mMin = new NumberParam(
              "Minimum distance",
              "The minimum distance to use for calculation",
              Field.MW.units.orElse(null),
              (Range<Double>) defaults.get(Field.MW).orElseThrow(),
              MAGNITUDE_DEFAULT_VALUES.lowerEndpoint());
          root.add(Key.M_MIN, context.serialize(mMin));

          @SuppressWarnings("unchecked")
          NumberParam mMax = new NumberParam(
              "Maximum magnitude",
              "The maximum magnitude to use for calculation",
              Field.MW.units.orElse(null),
              (Range<Double>) defaults.get(Field.MW).orElseThrow(),
              MAGNITUDE_DEFAULT_VALUES.upperEndpoint());
          root.add(Key.M_MAX, context.serialize(mMax));

          @SuppressWarnings("unchecked")
          NumberParam distance = new NumberParam(
              "Distance",
              "The distance between site and trace ",
              Field.RX.units.orElse(null),
              (Range<Double>) defaults.get(Field.RX).orElseThrow(),
              MAGNITUDE_DEFAULT_DISTANCE);
          root.add(Key.DISTANCE, context.serialize(distance));

          NumberParam step = new NumberParam(
              "Magnitude step",
              "The step between each magnitude ",
              null,
              Range.closed(0.0001, 1.0),
              MAGNITUDE_DEFAULT_STEP);
          root.add(Key.STEP, context.serialize(step));
        }

        /* Add only add those Gmms that belong to a Group. */
        List<Gmm> gmms = Arrays.stream(Gmm.Group.values())
            .flatMap(group -> group.gmms().stream())
            .sorted(Comparator.comparing(Object::toString))
            .distinct()
            .collect(Collectors.toList());

        GmmParam gmmParam = new GmmParam(
            GMM_NAME,
            GMM_INFO,
            gmms);
        root.add(Key.GMM, context.serialize(gmmParam));

        /* Add gmm groups. */
        GroupParam groups = new GroupParam(
            GROUP_NAME,
            GROUP_INFO,
            EnumSet.allOf(Gmm.Group.class));
        root.add(GROUP_KEY, context.serialize(groups));

        return root;
      }
    }
  }

  /*
   * Marker interface for spectra parameters. This was previously implemented as
   * an abstract class for label, info, and units, but Gson serialized subclass
   * fields before parent fields. To maintain a preferred order, one can write
   * custom serializers or repeat these four fields in each implementation.
   */
  private static interface Param {}

  @SuppressWarnings("unused")
  private static final class NumberParam implements Param {

    final String label;
    final String info;
    final String units;
    final Double min;
    final Double max;
    final Double value;

    NumberParam(GmmInput.Field field, Range<Double> constraint) {
      this(field, constraint, field.defaultValue);
    }

    NumberParam(GmmInput.Field field, Range<Double> constraint, Double value) {
      this.label = field.label;
      this.info = field.info;
      this.units = field.units.orElse(null);
      this.min = constraint.lowerEndpoint();
      this.max = constraint.upperEndpoint();
      this.value = Doubles.isFinite(value) ? value : null;
    }

    NumberParam(String label, String info, String units, Range<Double> constraint, Double value) {
      this.label = label;
      this.info = info;
      this.units = units;
      this.min = constraint.lowerEndpoint();
      this.max = constraint.upperEndpoint();
      this.value = Doubles.isFinite(value) ? value : null;
    }
  }

  @SuppressWarnings("unused")
  private static final class BooleanParam implements Param {

    final String label;
    final String info;
    final boolean value;

    BooleanParam(GmmInput.Field field) {
      this(field, field.defaultValue == 1.0);
    }

    BooleanParam(GmmInput.Field field, boolean value) {
      this.label = field.label;
      this.info = field.info;
      this.value = value;
    }
  }

  private static final String GMM_NAME = "Ground Motion Models";
  private static final String GMM_INFO = "Empirical models of ground motion";

  @SuppressWarnings("unused")
  private static class GmmParam implements Param {

    final String label;
    final String info;
    final List<Value> values;

    GmmParam(String label, String info, List<Gmm> gmms) {
      this.label = label;
      this.info = info;
      this.values = gmms.stream()
          .map(gmm -> new Value(gmm))
          .collect(Collectors.toList());
    }

    private static class Value {

      final String id;
      final String label;
      final ArrayList<String> supportedImts;
      final Constraints constraints;

      Value(Gmm gmm) {
        this.id = gmm.name();
        this.label = gmm.toString();
        this.supportedImts = supportedImts(gmm.supportedImts());
        this.constraints = gmm.constraints();
      }
    }

    private static ArrayList<String> supportedImts(Set<Imt> imts) {
      ArrayList<String> supportedImts = new ArrayList<>();

      for (Imt imt : imts) {
        supportedImts.add(imt.name());
      }

      return supportedImts;
    }

  }

  private static final String GROUP_KEY = "group";
  private static final String GROUP_NAME = "Ground Motion Model Groups";
  private static final String GROUP_INFO = "Groups of related ground motion models ";

  @SuppressWarnings("unused")
  private static final class GroupParam implements Param {

    final String label;
    final String info;
    final List<Value> values;

    GroupParam(String label, String info, Set<Gmm.Group> groups) {
      this.label = label;
      this.info = info;
      this.values = new ArrayList<>();
      for (Gmm.Group group : groups) {
        this.values.add(new Value(group));
      }
    }

    private static class Value {

      final String id;
      final String label;
      final List<Gmm> data;

      Value(Gmm.Group group) {
        this.id = group.name();
        this.label = group.toString();
        this.data = group.gmms();
      }
    }
  }
}
