package gov.usgs.earthquake.nshmp.www.gps;

import java.io.IOException;
import java.net.URL;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

import com.google.common.collect.ImmutableList;
import com.google.common.io.Resources;

import gov.usgs.earthquake.nshmp.data.DelimitedData;
import gov.usgs.earthquake.nshmp.data.DelimitedData.Record;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.json.Feature;
import gov.usgs.earthquake.nshmp.geo.json.FeatureCollection;
import gov.usgs.earthquake.nshmp.geo.json.GeoJson;
import gov.usgs.earthquake.nshmp.geo.json.GeoJson.Builder;
import gov.usgs.earthquake.nshmp.www.Utils;

/**
 * GPS data files must have recognized column headers. Some datasets may not
 * have all fields (e.g. vertical component, site ID and type, correlation).
 * Files may have comment lines starting with "#", these lines are skipped.
 * First three data lines of a sample file, space delimited here for ease of
 * reading:
 *
 * <pre>
 *  longitude  latitude      east     north   eastErr  northErr  correlation
 *  -118.1460   34.7540  -12.6980   15.8890    0.9600    0.8000    0.0010
 *  -119.9830   43.5900   -1.3320    2.0340    0.5900    0.6100   -0.0450
 * </pre>
 *
 * Other fields that may be included in the future: correction
 *
 * Target GeoJSON format:
 *
 * <pre>
 * {
 *   "type": "FeatureCollection",
 *   "features": [
 *     {
 *       "type": "Feature",
 *       "id": 0,             // record number or store stationID here
 *       "geometry": {
 *         "type": "Point",
 *         "coordinates": [-117.0, 34.0]
 *       },
 *       "properties": {
 *         "stationID": STATION_ID_STRING,  // stationID string, record number if field omitted
 *         "eastVelocity": 0.0000,
 *         "northVelocity": 0.0000,
 *         "verticalVelocity": 0.0000,
 *         "eastError": 0.0000,
 *         "northError": 0.0000,
 *         "verticalError": 0.0000,
 *         "correlation": 0.0000,
 *         "verticalPositiveIsDown": BOOLEAN,
 *         "velocityUnits": STRING  // e.g. cm/yr, mm/yr
 *       }
 *     }
 *   ]
 * }
 * </pre>
 */

/**
 * Container class to hold a {@code GpsData} {@code GpsDataSet}
 *
 * @author U.S. Geological Survey
 */
public class GpsData {

  // Map stationId to GpsRecord
  private final Map<String, GpsRecord> gpsData;

  // Set of GPS fields
  private final Set<String> gpsFields;

  // List of original lines in CSV file
  private final List<String> originalLines;

  private final String dataSource;

  private GpsData(
      Map<String, GpsRecord> gpsData,
      Set<String> gpsFields,
      List<String> originalLines,
      String dataSource) {
    this.gpsData = gpsData;
    this.gpsFields = gpsFields;
    this.originalLines = originalLines;
    this.dataSource = dataSource;
  }

  /**
   * Read {@code GpsData} from CSV file, converting ASCII records to
   * {@code GpsRecord}s.
   *
   * @return
   */
  public static GpsData readGpsData(GpsDataSet gpsDataSet) {

    String dataFile = Utils.GPS_RESOURCES + gpsDataSet.filename;

    URL url = Resources.getResource(GpsDataSet.class, Utils.GPS_RESOURCES + gpsDataSet.filename);

    // Read lines from resource
    List<String> originalLines;
    try {
      originalLines = Resources.readLines(url, StandardCharsets.UTF_8);
    } catch (IOException e) {
      e.printStackTrace();
      throw new RuntimeException("Could not read file [" + gpsDataSet.filename + "]");
    }

    // Create DelimitedData
    DelimitedData delimitedData = DelimitedData.comma(originalLines);

    Set<String> keys = delimitedData.columnKeys();

    Map<String, GpsRecord> gpsData;
    try {
      gpsData = buildGpsData(delimitedData, keys);
    } catch (IOException e) {
      e.printStackTrace();
      throw new RuntimeException("Problem reading delimited data in file [" + dataFile + "]");
    }

    return new GpsData(gpsData, keys, originalLines, dataFile);
  }

  private static Map<String, GpsRecord> buildGpsData(
      DelimitedData delimitedData,
      Set<String> fieldMap) throws IOException {
    // get list of gpsRecords
    List<Record> gpsRecords = delimitedData.records().collect(Collectors.toList());

    /*
     * Pass in record number for use as stationId, in case a stationId field is
     * not present. If stationId field does exist, use it as the gpsData key.
     *
     * Use LinkedHashMap to preserve record order
     */
    Map<String, GpsRecord> gpsData = new LinkedHashMap<>();
    IntStream.range(0, gpsRecords.size()).boxed()
        .forEach(i -> {
          String id = (fieldMap.contains(Field.STATIONID))
              ? gpsRecords.get(i).get(Field.STATIONID) : String.valueOf(i);
          GpsRecord rec =
              GpsRecord.fromDelimitedData(i, gpsRecords.get(i), fieldMap);
          gpsData.put(id, rec);
        });
    return Collections.unmodifiableMap(gpsData);

  }

  private static Optional<String> extractProperty(String line) {
    if (line.length() == 0 || !line.contains(":")) {
      return Optional.empty();
    }
    String prop = line.substring(line.lastIndexOf(":") + 1).strip();
    return (prop.length() == 0) ? Optional.empty() : Optional.of(prop);
  }

  /**
   * Get list of {@code Field} of ASCII input file.
   *
   * @return {@code List<Field>}
   */
  List<String> getFields() {
    return new ArrayList<>(this.gpsFields);
  }

  String getDataSource() {
    return new String(dataSource);
  }

  /**
   * Return copy of {@code List<String>} containing original lines from CSV
   * file.
   */
  List<String> originalLines() {
    return ImmutableList.copyOf(originalLines);
  }

  /**
   * Create FeatureCollection containing contents of {@code GpsData}
   *
   * @return {@code FeatureCollection}
   */
  FeatureCollection asFeatureCollection() {
    Builder fcBuilder = GeoJson.builder();

    gpsData.entrySet().stream()
        .forEach(
            e -> fcBuilder.add(e.getValue().toFeature().id(e.getKey()).build()));
    return fcBuilder.build();
  }

  /*
   * GPS Field Strings
   */
  static class Field {
    static final String STATIONID = "stationId";
    static final String SITETYPE = "siteType";
    static final String LONGITUDE = "lon";
    static final String LATITUDE = "lat";
    static final String VEL_E = "vel_E";
    static final String VEL_N = "vel_N";
    static final String VEL_V = "vel_V";
    static final String SIG_E = "sig_E";
    static final String SIG_N = "sig_N";
    static final String SIG_V = "sig_V";
    static final String CORRELATION = "correlation";
  }

  /**
   * Container class for a single {@code GpsRecord}
   */
  private static class GpsRecord {
    Location loc;
    String stationId;
    Map<String, Object> data;

    GpsRecord(Location loc, Map<String, Object> data) {
      this.loc = loc;
      this.data = data;
      this.stationId = (String) this.data.get(Field.STATIONID);
    }

    /**
     * Generate a {@code GpsRecord} from a delimited data {@code Record}.
     *
     * @param recNumber Record number in input file, used as {@code stationId}
     *        if this field is not present in {@code Record}
     * @param record Delimited data {@code Record}
     * @param keys {@code List<String>} of {@code Record} field keys
     * @return {@code GpsRecord}
     */
    static GpsRecord fromDelimitedData(
        int recNumber,
        Record record,
        Set<String> fieldMap) {
      double lat = Double.NaN;
      double lon = Double.NaN;
      boolean hasStationId = false;

      // Preserving Field order in record for now
      Map<String, Object> recordData = new LinkedHashMap<>();

      for (String field : fieldMap) {
        switch (field) {
          case Field.STATIONID:
            recordData.put(field, record.get(field));
            hasStationId = true;
            break;
          case Field.SITETYPE:
            recordData.put(field, record.get(field));
            break;
          case Field.LONGITUDE:
            lon = record.getDouble(field);
            break;
          case Field.LATITUDE:
            lat = record.getDouble(field);
            break;
          default:
            recordData.put(field, record.getDouble(field));
        }
      }

      if (!hasStationId) {
        recordData.put(Field.STATIONID, String.valueOf(recNumber));
      }

      return new GpsRecord(
          Location.create(lon, lat),
          Collections.unmodifiableMap(recordData));
    }

    @Override
    public String toString() {
      return String.format("<%s: %s>", this.loc.toString(), this.data.toString());
    }

    /**
     * Convert {@code GpsRecord} to GeoJSON feature
     *
     * @return GeoJSON feature builder
     */
    Feature.Builder toFeature() {
      Map<String, Object> props = new LinkedHashMap<>();

      data.keySet().stream()
          .filter(k -> !k.equals(Field.LATITUDE))
          .filter(k -> !k.equals(Field.LONGITUDE))
          .forEach(k -> props.put(k.toString(), data.get(k)));

      return Feature.point(this.loc)
          // TODO: stationId as Feature ID? Otherwise buried in properties
          // .id((String) props.get(Field.STATIONID))
          .properties(Collections.unmodifiableMap(props));
    }

  }

}
