package gov.usgs.earthquake.nshmp.www.gulf;

import static com.google.common.base.Preconditions.checkArgument;
import static com.google.common.base.Preconditions.checkNotNull;
import static com.google.common.base.Preconditions.checkState;

import java.io.IOException;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.net.URL;
import java.nio.charset.StandardCharsets;
import java.util.List;
import java.util.Map;
import java.util.Set;

import com.google.common.collect.ImmutableMap;
import com.google.common.io.Resources;

import gov.usgs.earthquake.nshmp.Maths;
import gov.usgs.earthquake.nshmp.data.DelimitedData;
import gov.usgs.earthquake.nshmp.data.DelimitedData.Record;
import gov.usgs.earthquake.nshmp.geo.BorderType;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.Region;
import gov.usgs.earthquake.nshmp.geo.Regions;
import gov.usgs.earthquake.nshmp.geo.json.Feature;
import gov.usgs.earthquake.nshmp.geo.json.GeoJson;
import gov.usgs.earthquake.nshmp.www.Utils;

/**
 * Read in Gulf Coast sediment depth data.
 *
 * @author U.S. Geological Survey
 */
public class GulfData {

  public static final double GULF_DATA_SPACING = 0.05;
  static final String GULF_REGION_FILENAME = "sediment-depth-region.geojson";
  static final String GULF_DATA_FILENAME = "CPADepthNSHM.csv";

  private final Region gulfRegion;

  private final ImmutableMap<Location, Double> gulfData;

  private GulfData(Region gulfRegion, ImmutableMap<Location, Double> gulfData) {
    this.gulfRegion = gulfRegion;
    this.gulfData = gulfData;
  }

  /**
   * Returns the {@code GulfData} associated with all files in the data
   * directory
   */
  public static GulfData readGulfData() {
    Feature feature = readGulfRegion();
    Region region = Regions.create(
        feature.properties().getString("title").orElseThrow(),
        feature.asPolygonBorder(),
        BorderType.MERCATOR_LINEAR);

    String dataFile = Utils.GULF_RESOURCES + GULF_DATA_FILENAME;
    ImmutableMap<Location, Double> gulfData = null;
    try {
      List<String> lines = Resources.readLines(
          Resources.getResource(GulfData.class, dataFile),
          StandardCharsets.UTF_8);
      DelimitedData data = DelimitedData.comma(lines);
      Set<String> keys = data.columnKeys();

      gulfData = data.records()
          .map(record -> GulfDataRecord.fromDelimitedRecord(record, keys))
          .collect(ImmutableMap.toImmutableMap(
              GulfDataRecord::getLocation,
              GulfDataRecord::getSedimentDepth));

    } catch (IOException e) {
      e.printStackTrace();
      throw new RuntimeException("Could not read [" + dataFile + "]");
    }

    return new GulfData(region, gulfData);
  }

  /*
   * Read Feature from GeoJSON
   */
  private static Feature readGulfRegion() {
    URL url = Resources.getResource(GulfData.class, Utils.GULF_RESOURCES + GULF_REGION_FILENAME);
    return GeoJson.from(url).toFeature();
  }

  /**
   * @return the gulfRegion
   */
  public Region getGulfRegion() {
    return gulfRegion;
  }

  // /**
  // * Returns all gulf data as {@code ImmutableMap<Location, Double>}
  // *
  // * @return the gulfData
  // */
  // public ImmutableMap<Location, Double> getGulfData() {
  // return gulfData;
  // }

  /*
   * Return sediment depth at target {@code Location}
   */
  public double getGulfData(Location loc) {
    loc = processLocation(loc);
    return checkNotNull(gulfData.get(loc));
  }

  /*
   * Return object containing grid point {@code Location} and {@code
   * sedimentDepth}
   */
  public GulfDataResponse getGulfDataResponse(Location loc) {
    loc = processLocation(loc);
    return new GulfDataResponse(loc, checkNotNull(gulfData.get(loc)));
  }

  /*
   * Return {@code sedimentDepth} as a {@code Feature}
   */
  public Feature getGulfDataFeature(Location loc) {
    loc = processLocation(loc);
    return Feature.point(loc)
        .properties(Map.of("sedimentDepth", gulfData.get(loc)))
        .build();
  }

  /*
   * Preprocess requested {@code Location}, validate against {@code gulfRegion},
   * round to grid, and validate against {@code gulfData}
   */
  private Location processLocation(Location loc) {
    checkArgument(gulfRegion.contains(loc),
        "Location [" + loc.toString() + "] is outside of Gulf data region");

    double lon = roundToGrid(loc.longitude, GULF_DATA_SPACING);
    double lat = roundToGrid(loc.latitude, GULF_DATA_SPACING);
    loc = Location.create(lon, lat);

    checkState(gulfData.containsKey(loc), "Location [%s] not found in Gulf dataset", loc);
    return loc;
  }

  /**
   * Round {@code value} to closest multiple of {@code spacing}
   *
   * TODO: clarify midpoint cases. It appears as if this case is handled
   * differently for negative (longitude) and positive (latitude) numbers (see
   * test for more details)
   *
   * TODO: move to www.Utils?
   *
   * @param value double
   * @param spacing double
   */
  public static double roundToGrid(double value, double spacing) {
    return Maths.round(
        Math.round(value / spacing) * spacing,
        BigDecimal.valueOf(spacing).scale(),
        RoundingMode.HALF_UP);
  }

  /**
   * Container class for a {@code Location} and {@code sedimentDepth}
   */
  static class GulfDataRecord {
    Location loc;
    double sedimentDepth;

    GulfDataRecord(Location loc, double depth) {
      this.loc = loc;
      this.sedimentDepth = depth;
    }

    /**
     * Returns the {@code Location}
     */
    Location getLocation() {
      return loc;
    }

    /**
     * Returns the {@code sedimentDepth}
     */
    double getSedimentDepth() {
      return sedimentDepth;
    }

    static GulfDataRecord fromDelimitedRecord(Record record, Set<String> keys) {
      double lon = Double.NaN;
      double lat = Double.NaN;
      double sedimentDepth = Double.NaN;

      for (String key : keys) {
        switch (key) {
          case "longitude":
            lon = record.getDouble(key);
            break;
          case "latitude":
            lat = record.getDouble(key);
            break;
          case "depth":
            sedimentDepth = record.getDouble(key);
            break;
          default:
            throw new RuntimeException("Key [" + key + "] not supported");
        }
      }

      return new GulfDataRecord(Location.create(lon, lat), sedimentDepth);
    }
  }

  /*
   * Container for gulf data snapped to grid
   */
  public static class GulfDataResponse {
    public final double longitude;
    public final double latitude;
    public final double sedimentDepth;

    public GulfDataResponse(Location loc, double sedimentDepth) {
      this.longitude = loc.longitude;
      this.latitude = loc.latitude;
      this.sedimentDepth = sedimentDepth;
    }

  }
}
