package gov.usgs.earthquake.nshmp.www.services;

import static gov.usgs.earthquake.nshmp.www.Utils.GSON;

import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import com.google.common.base.Enums;
import com.google.common.collect.Lists;
import com.google.common.collect.Range;

import gov.usgs.earthquake.nshmp.Text;
import gov.usgs.earthquake.nshmp.geo.json.GeoJson;
import gov.usgs.earthquake.nshmp.internal.UsRegion;
import gov.usgs.earthquake.nshmp.model.SourceFeature;
import gov.usgs.earthquake.nshmp.www.ResponseBody;
import gov.usgs.earthquake.nshmp.www.ResponseMetadata;
import gov.usgs.earthquake.nshmp.www.Utils;
import gov.usgs.earthquake.nshmp.www.Utils.Key;
import gov.usgs.earthquake.nshmp.www.WsUtils;
import gov.usgs.earthquake.nshmp.www.WsVersion;
import gov.usgs.earthquake.nshmp.www.fault.FaultGroup;
import gov.usgs.earthquake.nshmp.www.fault.FaultSections;
import gov.usgs.earthquake.nshmp.www.fault.NshmFaultSection;
import gov.usgs.earthquake.nshmp.www.meta.StringParameter;

import io.micronaut.http.HttpRequest;
import io.micronaut.http.HttpResponse;
import jakarta.inject.Singleton;

/**
 * Service handler for {@code FaultSectionsController}
 *
 * @author U.S. Geological Survey
 */
@Singleton
public class FaultSectionsService {

  private static final String SERVICE_NAME = "Fault Sections Service";
  private static final String SERVICE_DESCRIPTION = "Get fault sections as GeoJson";

  private static final List<String> SERVICE_SYNTAXES = List.of(
      "?group={group}&faults={NshmFaultSection}&raw={boolean}",
      "?id={int}&faults={NshmFaultSection}&raw={boolean}",
      "/{faultId:int}/{faultSections:string}/{rawGeoJson:boolean}",
      "/{group:String}/{faultSections:string}/{rawGeoJson:boolean}");

  private static Map<NshmFaultSection, FaultSections> FAULT_SECTIONS = null;

  public static HttpResponse<String> metadata(HttpRequest<?> request) {
    var metadata = GSON.toJson(getMetadata(request));
    return HttpResponse.ok(metadata);
  }

  public static HttpResponse<String> handleDoGetFaults(
      HttpRequest<?> request,
      List<String> group,
      List<Integer> id,
      List<String> name,
      NshmFaultSection faults,
      Boolean raw,
      Path path) {
    try {
      var query = new Query(group, id, name, raw, faults);
      setFaultSections(path);

      if (group == null && id == null && name == null) {
        return metadata(request);
      }

      var response = Text.cleanCoordinates(processRequest(request, query));
      return HttpResponse.ok(response);
    } catch (Exception e) {
      return Utils.handleError(e, SERVICE_NAME, request.getUri().getPath());
    }
  }

  public static HttpResponse<String> handleDoGetFaultsSlash(
      HttpRequest<?> request,
      String value,
      NshmFaultSection faults,
      Boolean raw,
      Path path) {
    List<String> groups = null;
    List<Integer> ids = null;

    try {
      ids = Lists.newArrayList(Integer.parseInt(value));
    } catch (NumberFormatException e) {
      // continue
    }

    if (value != null && ids == null) {
      if (Enums.getIfPresent(UsRegion.class, value).isPresent() ||
          Enums.getIfPresent(FaultGroup.class, value).isPresent()) {
        groups = Lists.newArrayList(value);
      } else {
        var msg = "Name search in slash delimited form not supported";
        var e = new Throwable(msg);
        return Utils.handleError(e, SERVICE_NAME, request.getUri().getPath());
      }
    }

    return handleDoGetFaults(request, groups, ids, null, faults, raw, path);
  }

  public static HttpResponse<String> handleDoPost(HttpRequest<?> request, Query query, Path path) {
    if (query == null) {
      query = new Query(null, null, null, false, null);
    }

    return handleDoGetFaults(
        request, query.groups, query.id, query.names, query.faults, query.raw, path);
  }

  static String processRequest(HttpRequest<?> request, Query query) {
    if (query.id != null) {
      return processRequestById(request, query.id, query.raw, query.faults);
    } else if (query.groups != null) {
      return processRequestByGroup(request, query.groups(), query.raw, query.faults);
    } else {
      return processRequestByName(request, query.names, query.raw, query.faults);
    }
  }

  static String processRequestByName(
      HttpRequest<?> request,
      List<String> names,
      boolean raw,
      NshmFaultSection faults) {
    var requestData = new RequestData<>(Key.NAME, names);
    var builder = GeoJson.builder();

    FAULT_SECTIONS.get(faults)
        .names(names)
        .forEach(fault -> builder.add(fault.source));

    if (raw) {
      return builder.build().toJson();
    } else {
      var response = ResponseBody.success()
          .name(SERVICE_NAME)
          .url(request.getUri().getPath())
          .metadata(new ResponseMetadata(WsVersion.appVersions()))
          .request(requestData)
          .response(builder.build().toJsonTree())
          .build();
      return GSON.toJson(response);
    }
  }

  static String processRequestById(
      HttpRequest<?> request,
      List<Integer> ids,
      boolean raw,
      NshmFaultSection faults) {
    var requestData = new RequestData<>(Key.ID, ids);
    GeoJson.Builder builder = GeoJson.builder();

    FAULT_SECTIONS.get(faults)
        .ids(ids)
        .forEach(fault -> builder.add(fault.source));

    if (raw) {
      return builder.build().toJson();
    } else {
      var response = ResponseBody.success()
          .name(SERVICE_NAME)
          .url(request.getUri().getPath())
          .metadata(new ResponseMetadata(WsVersion.appVersions()))
          .request(requestData)
          .response(builder.build().toJsonTree())
          .build();
      return GSON.toJson(response);
    }
  }

  static String processRequestByGroup(
      HttpRequest<?> request,
      List<GroupData> groups,
      boolean raw,
      NshmFaultSection faultSection) {
    var requestData = new RequestData<>(Key.GROUP, groups);
    var faults = new ArrayList<SourceFeature>();

    for (GroupData group : groups) {
      if (group.e instanceof UsRegion) {
        faults.addAll(FAULT_SECTIONS.get(faultSection).regions((UsRegion) group.e));
      } else if (group.e instanceof FaultGroup) {
        if (FaultGroup.ALL.equals(group.e)) {
          var ids = FAULT_SECTIONS.get(faultSection).ids().stream().collect(Collectors.toList());
          faults.addAll(FAULT_SECTIONS.get(faultSection).ids(ids));
        } else {
          faults.addAll(FAULT_SECTIONS.get(faultSection).ids(((FaultGroup) group.e).ids));
        }
      }
    }

    var builder = GeoJson.builder();
    faults.forEach(fault -> builder.add(fault.source));

    if (raw) {
      return builder.build().toJson();
    } else {
      var response = ResponseBody.success()
          .name(SERVICE_NAME)
          .url(request.getUri().getPath())
          .metadata(new ResponseMetadata(WsVersion.appVersions()))
          .request(requestData)
          .response(builder.build().toJsonTree())
          .build();
      return GSON.toJson(response);
    }
  }

  static List<GroupData> getGroups(List<String> groups) {
    return groups.stream()
        .map(String::toUpperCase)
        .map(group -> {
          if (Enums.getIfPresent(UsRegion.class, group).isPresent()) {
            return new GroupData(UsRegion.valueOf(group));
          } else if (Enums.getIfPresent(FaultGroup.class, group).isPresent()) {
            return new GroupData(FaultGroup.valueOf(group));
          }

          throw new RuntimeException("Group [" + group + "] not valid");
        }).collect(Collectors.toList());
  }

  private static void setFaultSections(Path path) {
    if (FAULT_SECTIONS == null) {
      FAULT_SECTIONS = FaultSections.getFaultSections(path);
    }
  }

  public static class Query {
    final List<String> groups;
    final List<Integer> id;
    final List<String> names;
    final boolean raw;
    final NshmFaultSection faults;

    Query(
        List<String> groups,
        List<Integer> id,
        List<String> names,
        Boolean raw,
        NshmFaultSection faults) {
      this.groups = groups;
      this.id = id;
      this.names = names;
      this.raw = raw == null ? false : raw;
      this.faults = faults == null ? NshmFaultSection.CONUS_2018 : faults;
    }

    List<GroupData> groups() {
      return groups == null ? null : getGroups(groups);
    }
  }

  public static class RequestData<T> {
    public final List<T> values;

    RequestData(Key key, List<T> values) {
      this.values = WsUtils.checkValue(key, values);
    }
  }

  public static class GroupData {
    private final transient Enum<?> e;
    public final String label;
    public final String id;

    GroupData(Enum<?> e) {
      this.e = e;
      label = e.toString();
      id = e.name();
    }
  }

  static ResponseBody<String, MetadataResponse> getMetadata(HttpRequest<?> request) {
    var url = request.getUri().getPath();
    return ResponseBody.<String, MetadataResponse> usage()
        .name(SERVICE_NAME)
        .url(url)
        .metadata(new ResponseMetadata(WsVersion.appVersions()))
        .request(url)
        .response(new MetadataResponse(request))
        .build();
  }

  private static Set<String> faultSectionGroups(NshmFaultSection faultSection) {
    var groups = new TreeSet<String>();

    var regions = FAULT_SECTIONS.get(faultSection).regions()
        .stream()
        .map(e -> e.name())
        .collect(Collectors.toSet());

    var faults = Stream.of(FaultGroup.values())
        .map(e -> e.name())
        .collect(Collectors.toSet());

    groups.addAll(regions);
    groups.addAll(faults);

    return groups;
  }

  @SuppressWarnings("unused")
  private static class MetadataResponse {
    final String description;
    final List<String> syntax;
    final Parameters parameters;

    MetadataResponse(HttpRequest<?> request) {
      var url = request.getUri().getPath();
      description = SERVICE_DESCRIPTION;
      syntax = SERVICE_SYNTAXES.stream()
          .map(syntax -> url + syntax)
          .collect(Collectors.toList());

      parameters = new Parameters();
    }
  }

  @SuppressWarnings("unused")
  private static class Parameters {
    final StringParameter group;
    final RangeParameter id;
    final StringParameter name;
    final ObjectParameter<FaultSectionParameter> faults;
    final ObjectParameter<Boolean> raw;

    Parameters() {
      var faultParams = Arrays.stream(NshmFaultSection.values())
          .map(faultSection -> new FaultSectionParameter(faultSection))
          .collect(Collectors.toSet());

      var groups = Arrays.stream(NshmFaultSection.values())
          .flatMap(faultSection -> faultSectionGroups(faultSection).stream())
          .collect(Collectors.toCollection(TreeSet::new));

      var ids = Arrays.stream(NshmFaultSection.values())
          .flatMap(faultSection -> FAULT_SECTIONS.get(faultSection).ids().stream())
          .collect(Collectors.toCollection(TreeSet::new));

      var names = Arrays.stream(NshmFaultSection.values())
          .flatMap(faultSection -> FAULT_SECTIONS.get(faultSection).names().stream())
          .collect(Collectors.toCollection(TreeSet::new));

      group = new StringParameter("Fault group", groups);
      id = new RangeParameter("Fault id range", Range.encloseAll(ids));
      name = new StringParameter("Fault name", names);

      faults = new ObjectParameter<FaultSectionParameter>(
          "NSHM fault Sections",
          faultParams);

      raw = new ObjectParameter<Boolean>(
          "Whether to return raw GeoJSON",
          Set.of(true, false));
    }
  }

  @SuppressWarnings("unused")
  private static class ObjectParameter<T> {
    final String label;
    final Set<T> values;

    ObjectParameter(String label, Set<T> values) {
      this.label = label;
      this.values = values;
    }
  }

  @SuppressWarnings("unused")
  private static class RangeParameter {
    final String label;
    final String range;

    RangeParameter(String label, Range<Integer> range) {
      this.label = label;
      this.range = range.toString();
    }
  }

  @SuppressWarnings("unused")
  private static class FaultSectionParameter {
    final NshmFaultSection value;
    final String display;
    final FaultSectionSupports supports;

    FaultSectionParameter(NshmFaultSection faultSection) {
      var faults = FAULT_SECTIONS.get(faultSection);

      value = faultSection;
      display = faultSection.label();
      supports = new FaultSectionSupports(
          faultSectionGroups(faultSection),
          faults.names(),
          Range.encloseAll(faults.ids()));
    }
  }

  @SuppressWarnings("unused")
  private static class FaultSectionSupports {
    final Set<String> group;
    final String id;
    final Set<String> name;

    FaultSectionSupports(Set<String> group, Set<String> name, Range<Integer> range) {
      this.group = group;
      this.id = range.toString();
      this.name = name;
    }
  }

}
