package gov.usgs.earthquake.nshmp.www.services;

import static gov.usgs.earthquake.nshmp.www.Utils.GSON;
import static gov.usgs.earthquake.nshmp.www.WsUtils.checkValue;

import java.util.EnumSet;

import gov.usgs.earthquake.nshmp.www.ResponseBody;
import gov.usgs.earthquake.nshmp.www.ResponseMetadata;
import gov.usgs.earthquake.nshmp.www.Utils;
import gov.usgs.earthquake.nshmp.www.Utils.Key;
import gov.usgs.earthquake.nshmp.www.WsVersion;
import gov.usgs.earthquake.nshmp.www.gps.GpsDataSet;
import gov.usgs.earthquake.nshmp.www.gps.GpsDataSets;
import gov.usgs.earthquake.nshmp.www.meta.EnumParameter;

import io.micronaut.http.HttpRequest;
import io.micronaut.http.HttpResponse;
import jakarta.inject.Singleton;

/**
 * Service handler for {@code GpsController}
 *
 * @author U.S. Geological Survey
 */
@Singleton
public class GpsService {

  private static GpsDataSets GPSDATASETS = null;
  private static final String SERVICE_NAME = "GPS Data Service";
  private static final String SERVICE_DESCRIPTION =
      "Get GPS NSHM data, velocity units are mm/yr and positive up";
  private static final String SERVICE_SYNTAX = "?model={model}&format={format}";

  public static HttpResponse<String> metadata(HttpRequest<?> request) {
    var metadata = GSON.toJson(getMetadata(request));
    return HttpResponse.ok(metadata);
  }

  public static HttpResponse<String> handleDoGetGps(
      HttpRequest<?> request,
      GpsDataSet model,
      GpsService.GpsFormat format) {
    try {
      setGps();

      if (model == null) {
        var metadata = GSON.toJson(getMetadata(request));
        return HttpResponse.ok(metadata);
      }

      var query = new Query(model, format);
      checkValue(Key.GPSDATASET, model);
      var response = processGpsRequest(request, query);
      // appending .contentType(MediaType.TEXT_PLAIN_TYPE) can preserve JSON
      // pretty formatting
      return HttpResponse.ok(response);
    } catch (Exception e) {
      return Utils.handleError(e, SERVICE_NAME, request.getUri().getPath());
    }
  }

  static String processGpsRequest(HttpRequest<?> request, Query query) {
    RequestData requestData = buildRequest(query);

    if (query.format == GpsFormat.RAW_JSON) {
      return GPSDATASETS.getJsonString(query.model);
    }
    if (query.format == GpsFormat.CSV) {
      return GPSDATASETS.getCsv(requestData.model);
    }
    var values = GPSDATASETS.getJsonTree(requestData.model);

    var response = ResponseBody.success()
        .name(SERVICE_NAME)
        .url(request.getUri().getPath())
        .metadata(new ResponseMetadata(WsVersion.appVersions()))
        .request(requestData)
        .response(values)
        .build();
    return GSON.toJson(response, ResponseBody.class);
  }

  private static RequestData buildRequest(Query query) {
    return new RequestData(query.model, query.format);
  }

  private static void setGps() {
    if (GPSDATASETS == null) {
      GPSDATASETS = GpsDataSets.loadGpsDataSets();
    }

  }

  public static class Query {
    public final GpsDataSet model;
    public final GpsService.GpsFormat format;

    public Query(GpsDataSet model, GpsService.GpsFormat format) {
      this.model = model;
      this.format = format == null ? GpsFormat.CSV : format;
    }
  }

  public static class RequestData {
    public final GpsDataSet model;
    public final GpsFormat format;

    RequestData(GpsDataSet model, GpsFormat format) {
      this.model = model;
      this.format = format;
    }

  }

  public enum GpsFormat {
    CSV,
    JSON,
    RAW_JSON;
  }

  static ResponseBody<String, MetadataResponse> getMetadata(HttpRequest<?> request) {
    var url = request.getUri().getPath();
    return ResponseBody.<String, MetadataResponse> usage()
        .name(SERVICE_NAME)
        .url(url)
        .metadata(new ResponseMetadata(WsVersion.appVersions()))
        .request(url)
        .response(new MetadataResponse(request))
        .build();
  }

  public static class MetadataResponse {
    final String description;
    final String syntax;
    final EnumParameter<GpsDataSet> datasets;
    final EnumParameter<GpsFormat> formats;

    public MetadataResponse(HttpRequest<?> request) {
      description = SERVICE_DESCRIPTION;
      syntax = request.getUri().getPath() + SERVICE_SYNTAX;
      datasets = new EnumParameter<>(
          "GPS Dataset",
          EnumSet.allOf(GpsDataSet.class));
      formats = new EnumParameter<>(
          "Formats",
          EnumSet.allOf(GpsFormat.class));
    }
  }
}
