package gov.usgs.earthquake.nshmp.www.services;

import static gov.usgs.earthquake.nshmp.www.Utils.GSON;
import static gov.usgs.earthquake.nshmp.www.WsUtils.checkValue;

import com.google.gson.JsonElement;

import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.Region;
import gov.usgs.earthquake.nshmp.www.ResponseBody;
import gov.usgs.earthquake.nshmp.www.ResponseMetadata;
import gov.usgs.earthquake.nshmp.www.Utils;
import gov.usgs.earthquake.nshmp.www.Utils.Key;
import gov.usgs.earthquake.nshmp.www.WsVersion;
import gov.usgs.earthquake.nshmp.www.gulf.GulfData;
import gov.usgs.earthquake.nshmp.www.gulf.GulfData.GulfDataResponse;

import io.micronaut.http.HttpRequest;
import io.micronaut.http.HttpResponse;
import jakarta.inject.Singleton;

@Singleton
public class GulfService {

  private static GulfData GULF_DATA = null;
  private static final String SERVICE_NAME = "Gulf Coast Sediment Depth Service";
  private static final String SERVICE_DESCRIPTION = "Get gulf coast sediment depth";
  private static final String SERVICE_SYNTAX =
      "?longitude={longitude}&latitude={latitude}";

  public static HttpResponse<String> metadata(HttpRequest<?> request) {
    String metadata = GSON.toJson(getMetadata(request));
    return HttpResponse.ok(metadata);
  }

  private static void setGulf() {
    if (GULF_DATA == null) {
      GULF_DATA = GulfData.readGulfData();
    }
  }

  private static ResponseBody<String, MetadataResponse> getMetadata(HttpRequest<?> request) {
    var url = request.getUri().getPath();
    return ResponseBody.<String, MetadataResponse> usage()
        .name(SERVICE_NAME)
        .url(url)
        .metadata(new ResponseMetadata(WsVersion.appVersions()))
        .request(url)
        .response(new MetadataResponse(request))
        .build();
  }

  public static HttpResponse<String> handleDoGetGulf(
      HttpRequest<?> request,
      Double longitude,
      Double latitude) {
    try {
      setGulf();

      if (longitude == null && latitude == null) {
        var metadata = GSON.toJson(getMetadata(request));
        return HttpResponse.ok(metadata);
      }

      checkValue(Key.LONGITUDE, longitude);
      checkValue(Key.LATITUDE, latitude);

      var query = new Query(longitude, latitude, false);

      String response = GSON.toJson(processGulf(request, query), ResponseBody.class);

      return HttpResponse.ok(response);

    } catch (Exception e) {
      return Utils.handleError(e, SERVICE_NAME, request.getUri().getPath());
    }
  }

  private static ResponseBody<RequestData, JsonElement> processGulf(HttpRequest<?> request,
      Query query) {
    var requestData = buildRequest(query, GulfData.GULF_DATA_SPACING);

    var loc = Location.create(requestData.longitude, requestData.latitude);

    // The request includes the original coordinates, no rounding is performed
    // here, only in the GulfData class, so the coordinates in the response
    // match the grid and returned data.
    var value = GSON.toJsonTree(GULF_DATA.getGulfDataResponse(loc), GulfDataResponse.class);

    // return point as Feature
    // var value = GSON.toJsonTree(GULF_DATA.getGulfDataFeature(loc),
    // Feature.class);

    return ResponseBody.<RequestData, JsonElement> success()
        .name(SERVICE_NAME)
        .url(request.getUri().getPath())
        .metadata(new ResponseMetadata(WsVersion.appVersions()))
        .request(requestData)
        .response(value)
        .build();
  }

  /*
   * Build request using original longitude, latitude coordinates. Rounding to
   * the grid is handled by {@code GulfData}.
   */
  private static RequestData buildRequest(Query query, double spacing) {
    return new RequestData(query.longitude, query.latitude);
  }

  public static class MetadataResponse {
    final String description;
    final String syntax;
    final Region gulfRegion;

    public MetadataResponse(HttpRequest<?> request) {
      description = SERVICE_DESCRIPTION;
      syntax = request.getUri().getPath() + SERVICE_SYNTAX;
      gulfRegion = GULF_DATA.getGulfRegion();
    }
  }

  public static class Query {
    public final Double longitude;
    public final Double latitude;
    public final Boolean raw;

    public Query(Double longitude, Double latitude, Boolean raw) {
      this.longitude = longitude;
      this.latitude = latitude;
      this.raw = raw;
    }
  }

  public static class RequestData {
    public final double longitude;
    public final double latitude;

    RequestData(double longitude, double latitude) {
      this.longitude = longitude;
      this.latitude = latitude;
    }
  }
}
