package gov.usgs.earthquake.nshmp.www;

import static gov.usgs.earthquake.nshmp.www.Utils.GSON;
import static org.junit.jupiter.api.Assertions.assertEquals;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.lang.reflect.Type;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;
import org.junit.jupiter.params.provider.MethodSource;

import com.google.common.collect.Lists;
import com.google.common.reflect.TypeToken;
import com.google.gson.JsonElement;

import gov.usgs.earthquake.nshmp.geo.json.Feature;
import gov.usgs.earthquake.nshmp.geo.json.FeatureCollection;
import gov.usgs.earthquake.nshmp.geo.json.GeoJson;
import gov.usgs.earthquake.nshmp.internal.UsRegion;
import gov.usgs.earthquake.nshmp.model.SourceFeature;
import gov.usgs.earthquake.nshmp.www.fault.FaultGroup;
import gov.usgs.earthquake.nshmp.www.fault.FaultSections;
import gov.usgs.earthquake.nshmp.www.fault.NshmFaultSection;
import gov.usgs.earthquake.nshmp.www.meta.Status;
import gov.usgs.earthquake.nshmp.www.services.FaultSectionsService.GroupData;
import gov.usgs.earthquake.nshmp.www.services.FaultSectionsService.RequestData;

import io.micronaut.http.HttpRequest;
import io.micronaut.test.extensions.junit5.annotation.MicronautTest;
import jakarta.inject.Inject;

@MicronautTest
@SuppressWarnings("serial")
public class FaultSectionsControllerTests {

  private static NshmFaultSection FAULT_SECTION = NshmFaultSection.CONUS_2018;
  private static Path PATH = Paths.get(Utils.FAULT_RESOURCES);
  private static FaultSections FAULT_SECTIONS = FaultSections.getFaultSections(PATH)
      .get(FAULT_SECTION);

  private HttpRequest<?> request;

  @Inject
  FaultSectionsController controller;

  @BeforeEach
  public final void beforeEach() {
    request = HttpRequest.GET("");
  }

  @ParameterizedTest(name = "{index} ==> Fault Name: {0}")
  @MethodSource("byNames")
  public final void doGetByNamesTests(String name) throws UnsupportedEncodingException {
    List<SourceFeature> faultSections = FAULT_SECTIONS.names(Lists.newArrayList(name));

    var names = List.of(name);

    // Test query call
    var http = controller.doGetFaults(request, null, null, names, FAULT_SECTION, false);
    testResponseByName(http.body(), faultSections, name);

    // Test query call with raw flag
    var httpRaw = controller.doGetFaults(request, null, null, names, FAULT_SECTION, true);
    testRawResponse(httpRaw.body(), faultSections);
  }

  @ParameterizedTest(name = "{index} ==> Fault Id: {0}")
  @MethodSource("byId")
  public final void doGetByIdsTests(int id) throws IOException {
    List<SourceFeature> faultSections = FAULT_SECTIONS.ids(Lists.newArrayList(id));

    var ids = List.of(id);

    // Test query call
    var http = controller.doGetFaults(request, null, ids, null, FAULT_SECTION, false);
    testResponseById(http.body(), faultSections, id);

    // Test query call with raw flag
    var httpRaw = controller.doGetFaults(request, null, ids, null, FAULT_SECTION, true);
    testRawResponse(httpRaw.body(), faultSections);

    // Test slash delimited
    var httpSlash = controller.doGetFaultsSlash(
        request, Integer.toString(id), FAULT_SECTION, false);
    testResponseById(httpSlash.body(), faultSections, id);

    // Test slash delimited with raw flag
    var httpRawSlash = controller.doGetFaultsSlash(
        request, Integer.toString(id), FAULT_SECTION, true);
    testRawResponse(httpRawSlash.body(), faultSections);
  }

  @ParameterizedTest(name = "{index} ==> Region: {0}")
  @EnumSource(value = UsRegion.class)
  public final void doGetByRegionTests(UsRegion region) {
    testGroupCall(region);
  }

  @ParameterizedTest(name = "{index} ==> Group: {0}")
  @EnumSource(value = FaultGroup.class)
  public final void doGetByGroupTests(FaultGroup group) {
    testGroupCall(group);
  }

  @Test
  public final void doGetNameDelimitedIAE() {
    var http = controller.doGetFaultsSlash(request, "Some bad fault name", FAULT_SECTION, false);
    Type type = new TypeToken<ResponseBody<String, String>>(
        ResponseBody.class) {}.getType();
    ResponseBody<String, JsonElement> res = GSON.fromJson(http.body(), type);
    assertEquals(Status.ERROR, res.getStatus());
  }

  private <E extends Enum<E>> void testGroupCall(E group) {
    try {
      var faultSections = group instanceof UsRegion
          ? FAULT_SECTIONS.regions((UsRegion) group) : FAULT_SECTIONS.ids(((FaultGroup) group).ids);

      if (FaultGroup.ALL.equals(group)) {
        faultSections = FAULT_SECTIONS.ids(
            FAULT_SECTIONS.ids().stream().collect(Collectors.toList()));
      }

      var groups = List.of(group.name());

      // Test query call
      var http = controller.doGetFaults(request, groups, null, null, FAULT_SECTION, false);
      testResponseByGroup(http.body(), faultSections, group);

      // Test query call with raw flag
      var httpRaw = controller.doGetFaults(request, groups, null, null, FAULT_SECTION, true);
      testRawResponse(httpRaw.body(), faultSections);

      // Test slash delimited
      var httpSlash = controller.doGetFaultsSlash(request, group.name(), FAULT_SECTION, false);
      testResponseByGroup(httpSlash.body(), faultSections, group);

      // Test slash delimited with raw flag
      var httpRawSlash = controller.doGetFaultsSlash(request, group.name(), FAULT_SECTION, true);
      testRawResponse(httpRawSlash.body(), faultSections);
    } catch (Exception e) {
      if (!e.getMessage().contains("Region") && !e.getMessage().contains("not found")) {
        throw new RuntimeException(e);
      }
    }
  }

  private <E extends Enum<E>> void testResponseByGroup(
      String svcResponse,
      List<SourceFeature> faultSections,
      E group) {
    Type type = new TypeToken<ResponseBody<RequestData<GroupData>, JsonElement>>(
        ResponseBody.class) {}.getType();

    ResponseBody<RequestData<GroupData>, JsonElement> res = GSON.fromJson(svcResponse, type);

    assertEquals(Status.SUCCESS, res.getStatus());
    assertEquals(group.toString(), res.getRequest().values.get(0).label);
    assertEquals(group.name(), res.getRequest().values.get(0).id);
    FeatureCollection fc = GeoJson.from(res.getResponse().toString()).toFeatureCollection();
    testRawResponse(faultSections, fc);

    if (group instanceof UsRegion) {
      faultSections.forEach(faultSection -> {
        if (!group.name().equals(faultSection.state.orElseThrow())) {
          assertEquals(group.name(), faultSection.states.get(1));
        } else {
          assertEquals(group.name(), faultSection.state.orElseThrow());
        }
      });
    }
  }

  private void testResponseById(
      String svcResponse,
      List<SourceFeature> faultSections,
      int id) {
    Type type =
        new TypeToken<ResponseBody<RequestData<Integer>, JsonElement>>(ResponseBody.class) {}
            .getType();
    ResponseBody<RequestData<Integer>, JsonElement> res = GSON.fromJson(svcResponse, type);

    assertEquals(Status.SUCCESS, res.getStatus());
    assertEquals(id, faultSections.get(0).id);
    assertEquals(id, res.getRequest().values.get(0));
    FeatureCollection fc = GeoJson.from(res.getResponse().toString()).toFeatureCollection();
    testRawResponse(faultSections, fc);
  }

  private void testResponseByName(
      String svcResponse,
      List<SourceFeature> faultSections,
      String name) {
    Type type = new TypeToken<ResponseBody<RequestData<String>, JsonElement>>(ResponseBody.class) {}
        .getType();
    ResponseBody<RequestData<String>, JsonElement> res = GSON.fromJson(svcResponse, type);

    assertEquals(Status.SUCCESS, res.getStatus());
    assertEquals(name,
        faultSections.get(0).name + " - " + faultSections.get(0).state.orElseThrow());
    assertEquals(name, res.getRequest().values.get(0));
    FeatureCollection fc = GeoJson.from(res.getResponse().toString()).toFeatureCollection();
    testRawResponse(faultSections, fc);
  }

  private void testRawResponse(List<SourceFeature> faultSections, FeatureCollection fc) {
    assertEquals(faultSections.size(), fc.features().size());
    List<Feature> features = fc.features();

    for (int i = 0; i < faultSections.size(); i++) {
      testRawResponse(faultSections.get(i), features.get(i));
    }
  }

  private void testRawResponse(String svcResponse, List<SourceFeature> faultSections) {
    FeatureCollection fc = GeoJson.from(svcResponse).toFeatureCollection();
    testRawResponse(faultSections, fc);
  }

  private void testRawResponse(SourceFeature faultSection, Feature feature) {
    Feature expected = faultSection.source;

    switch (feature.type()) {
      case LINE_STRING:
        assertEquals(expected.asLineString(), feature.asLineString());
        break;
      case MULTI_LINE_STRING:
        assertEquals(expected.asMultiLineString(), feature.asMultiLineString());
        break;
      default:
        throw new IllegalStateException("Geometry type [" + feature.type() + "] not supported");
    }
  }

  private static Stream<String> byNames() {
    return FAULT_SECTIONS.names().stream();
  }

  private static Stream<Integer> byId() {
    return FAULT_SECTIONS.ids().stream();
  }

}
