package gov.usgs.earthquake.nshmp.www;

import static gov.usgs.earthquake.nshmp.www.Utils.GSON;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertNotEquals;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.junit.jupiter.api.Assertions.fail;

import java.lang.reflect.Type;

import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;

import com.google.common.reflect.TypeToken;
import com.google.gson.JsonElement;

import gov.usgs.earthquake.nshmp.geo.json.Feature;
import gov.usgs.earthquake.nshmp.geo.json.FeatureCollection;
import gov.usgs.earthquake.nshmp.geo.json.GeoJson;
import gov.usgs.earthquake.nshmp.geo.json.Properties;
import gov.usgs.earthquake.nshmp.www.gps.GpsDataSet;
import gov.usgs.earthquake.nshmp.www.gps.GpsDataSets;
import gov.usgs.earthquake.nshmp.www.meta.Status;
import gov.usgs.earthquake.nshmp.www.services.GpsService;
import gov.usgs.earthquake.nshmp.www.services.GpsService.GpsFormat;
import gov.usgs.earthquake.nshmp.www.services.GpsService.RequestData;

import io.micronaut.http.HttpRequest;
import io.micronaut.test.extensions.junit5.annotation.MicronautTest;
import jakarta.inject.Inject;

@MicronautTest
class GpsControllerTests {

  private static final GpsDataSets DATA = GpsDataSets.loadGpsDataSets();

  @SuppressWarnings("serial")
  private static final Type REQUEST_DATA_TYPE =
      new TypeToken<ResponseBody<RequestData, JsonElement>>(ResponseBody.class) {}.getType();

  @Inject
  GpsController controller;

  @ParameterizedTest
  @EnumSource(value = GpsDataSet.class)
  final void gpsDataTests(GpsDataSet model) {
    var request = HttpRequest.GET("");

    // loop over available formats
    for (GpsFormat format : GpsService.GpsFormat.values()) {
      var http = controller.doGetGpsData(request, model, format);
      httpTests(http.body(), model, format);
    }
  }

  private static void httpTests(String http, GpsDataSet model, GpsFormat format) {
    if (format.equals(GpsFormat.JSON)) {
      ResponseBody<RequestData, JsonElement> response = GSON.fromJson(http, REQUEST_DATA_TYPE);
      var actual = GeoJson.from(response.getResponse().toString()).toFeatureCollection();

      assertEquals(Status.SUCCESS, response.getStatus());
      assertTrue(Status.SUCCESS.equals(response.getStatus()));
      assertNotEquals(Status.ERROR, response.getStatus());

      testDataResponse(model, actual);

    } else if (format.equals(GpsFormat.RAW_JSON)) {
      var actual = GeoJson.from(http).toFeatureCollection();

      testDataResponse(model, actual);

    } else if (format.equals(GpsFormat.CSV)) {
      var actual = http;
      var expected = DATA.getCsv(model);

      assertEquals(expected, actual);
      assertTrue(expected.equals(actual));

    } else {
      fail("Unexpected GpsFormat [" + format.name() + "]");
    }
  }

  private static void testDataResponse(GpsDataSet model, FeatureCollection actual) {
    var features = actual.features();

    var expected = GeoJson.from(DATA.getJsonString(model)).toFeatureCollection().features();

    assertEquals(expected.size(), features.size());

    for (int i = 0; i < features.size(); i++) {

      Feature actualFeature = features.get(i);
      Properties actualProps = actualFeature.properties();

      Feature expectedFeature = expected.get(i);
      Properties expectedProps = expectedFeature.properties();

      assertEquals(expectedFeature.asPoint().longitude, actualFeature.asPoint().longitude);
      assertEquals(expectedFeature.asPoint().latitude, actualFeature.asPoint().latitude);
      assertEquals(expectedFeature.asPoint(), actualFeature.asPoint());

      assertEquals(expectedProps.map(), actualProps.map());
      testProperties(expectedProps, actualProps);

    }

  }

  private static void testProperties(Properties expected, Properties actual) {
    assertEquals(expected.map().keySet().size(), actual.map().keySet().size());
    for (String field : expected.map().keySet()) {
      switch (field) {
        case "lon":
        case "lat":
          break;
        case "stationId":
        case "siteType":
          if (expected.getString(field).isPresent()) {
            assertTrue(actual.getString(field).isPresent());
            assertEquals(
                expected.getString(field).orElseThrow(),
                actual.getString(field).orElseThrow());
          } else {
            assertTrue(actual.getString(field).isEmpty());
          }
        default:
          if (expected.getDouble(field).isPresent()) {
            assertTrue(actual.getDouble(field).isPresent());
            assertEquals(expected.getDouble(field).getAsDouble(),
                actual.getDouble(field).getAsDouble());
          } else {
            assertTrue(actual.getDouble(field).isEmpty());
          }
      }
    }
  }

}
