package gov.usgs.earthquake.nshmp.www.fault;

import static org.junit.jupiter.api.Assertions.assertDoesNotThrow;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertTrue;

import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Stream;

import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.EnumSource;
import org.junit.jupiter.params.provider.MethodSource;

import com.google.common.collect.Lists;
import com.google.common.collect.Range;

import gov.usgs.earthquake.nshmp.internal.UsRegion;
import gov.usgs.earthquake.nshmp.www.Utils;

class FaultSectionsTests {

  private static final Path PATH = Paths.get(Utils.FAULT_RESOURCES);
  private static final FaultSections FAULT_SECTIONS = FaultSections
      .getFaultSections(PATH)
      .get(NshmFaultSection.CONUS_2018);
  private static Map<String, Range<Integer>> ID_RANGES = new HashMap<>();

  static {
    ID_RANGES.put(UsRegion.CA.name(), Range.closed(1, 999));
    ID_RANGES.put(UsRegion.NV.name(), Range.closed(1000, 1999));
    ID_RANGES.put(UsRegion.AZ.name(), Range.closed(2000, 2099));
    ID_RANGES.put(UsRegion.CO.name(), Range.closed(2100, 2199));
    ID_RANGES.put(UsRegion.ID.name(), Range.closed(2200, 2299));
    ID_RANGES.put(UsRegion.MT.name(), Range.closed(2300, 2399));
    ID_RANGES.put(UsRegion.NM.name(), Range.closed(2400, 2499));
    ID_RANGES.put(UsRegion.OR.name(), Range.closed(2500, 2599));
    ID_RANGES.put(UsRegion.TX.name(), Range.closed(2600, 2699));
    ID_RANGES.put(UsRegion.UT.name(), Range.closed(2700, 2799));
    ID_RANGES.put(UsRegion.WA.name(), Range.closed(2800, 2899));
    ID_RANGES.put(UsRegion.WY.name(), Range.closed(2900, 2999));
    ID_RANGES.put(UsRegion.AK.name(), Range.closed(4000, 4999));
    ID_RANGES.put(UsRegion.HI.name(), Range.closed(5000, 5999));
    ID_RANGES.put(UsRegion.AS.name(), Range.closed(7000, 7099));
    ID_RANGES.put(UsRegion.GU.name(), Range.closed(7100, 7199));
    ID_RANGES.put(UsRegion.MP.name(), Range.closed(7100, 7199));
    ID_RANGES.put(UsRegion.PR.name(), Range.closed(7200, 7299));
    ID_RANGES.put(UsRegion.VI.name(), Range.closed(7200, 7299));
    ID_RANGES.put(Group.CASCADIA.name(), Range.closed(3100, 3199));
    ID_RANGES.put(Group.NEW_MADRID.name(), Range.closed(3000, 3099));
    ID_RANGES.put(Group.MEERS.name(), Range.closed(3200, 3999));
  }

  @Test
  final void faultSectionLoadTests() {
    assertDoesNotThrow(() -> FaultSections.getFaultSections(PATH));
  }

  @ParameterizedTest(name = "{index} ==> Region: {0}")
  @MethodSource("byRegion")
  final void regionTests(UsRegion region) {
    var faultSections = FAULT_SECTIONS.regions(region);

    faultSections.forEach(faultSection -> {
      if (!region.name().equals(faultSection.state.orElseThrow())) {
        assertEquals(region.name(), faultSection.states.get(1));
      } else {
        assertEquals(region.name(), faultSection.state.orElseThrow());
      }
    });
  }

  @ParameterizedTest(name = "{index} ==> Group: {0}")
  @EnumSource(value = FaultGroup.class)
  final void groupsTests(FaultGroup group) {
    var faultSections = FAULT_SECTIONS.ids(group.ids);
    assertEquals(group.ids.size(), faultSections.size());

    for (int i = 0; i < group.ids.size(); i++) {
      assertEquals(group.ids.get(i), faultSections.get(i).id);
    }
  }

  @ParameterizedTest(name = "{index} ==> Id: {0}")
  @MethodSource("byId")
  final void idsTests(int id) {
    var section = FAULT_SECTIONS.ids(Lists.newArrayList(id)).get(0);
    assertEquals(id, section.id);

    if (section.name.startsWith("New Madrid")) {
      assertTrue(ID_RANGES.get(Group.NEW_MADRID.name()).contains(section.id));
    } else if (section.name.startsWith("Cascadia")) {
      assertTrue(ID_RANGES.get(Group.CASCADIA.name()).contains(section.id));
    } else if (section.name.startsWith("Meers")) {
      assertTrue(ID_RANGES.get(Group.MEERS.name()).contains(section.id));
    } else {
      var state = UsRegion.valueOf(section.state.orElseThrow()).name();
      assertTrue(ID_RANGES.get(state).contains(section.id));
    }
  }

  @ParameterizedTest(name = "{index} ==> Name: {0}, State: {1}")
  @MethodSource("byName")
  final void namesTests(String name, String state) {
    var section = FAULT_SECTIONS.names(List.of(name)).get(0);

    assertEquals(String.format("%s - %s", section.name, section.state.orElseThrow()), name);
    assertEquals(section.state.orElseThrow(), state);
  }

  private static Stream<Arguments> byId() {
    return FAULT_SECTIONS.ids().stream()
        .map(Arguments::of);
  }

  private static Stream<Arguments> byName() {
    return FAULT_SECTIONS.names().stream()
        .map(name -> nameSplit(name));
  }

  private static Stream<Arguments> byRegion() {
    return FAULT_SECTIONS.regions().stream()
        .map(Arguments::of);
  }

  private static Arguments nameSplit(String name) {
    var split = name.split("-");
    var state = split[split.length - 1].trim();
    return Arguments.of(name, state);
  }

  private static enum Group {
    NEW_MADRID,
    CASCADIA,
    MEERS;
  }

}
