package gov.usgs.earthquake.nshmp.www.gps;

import static org.junit.jupiter.api.Assertions.assertEquals;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;

import com.google.common.io.Resources;

import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.json.Feature;
import gov.usgs.earthquake.nshmp.geo.json.GeoJson;
import gov.usgs.earthquake.nshmp.www.Utils;
import gov.usgs.earthquake.nshmp.www.gps.GpsData.Field;

class GpsDataSetsTests {

  private static final GpsDataSets dataSets = GpsDataSets.loadGpsDataSets();

  private static final Map<GpsDataSet, Map<Object, Object>> expected;

  private static List<String> propertiesToCheck =
      Arrays.asList("UNITS", "POSITIVE_UP", Field.SITETYPE, "NOT_A_PROPERTY");
  private static List<String> doublesToCheck =
      Arrays.asList(Field.VEL_E, Field.SIG_N, Field.VEL_V, Field.CORRELATION);

  static {
    Map<GpsDataSet, Map<Object, Object>> expectedMap = new HashMap<>();

    // AK_2020
    expectedMap.put(GpsDataSet.AK_2020, Map.of(
        "count", 321,
        "first", Map.of(
            "id", "2437",
            "coords", Location.create(-135.44, 58.42),
            Field.SITETYPE, "Campaign",
            Field.VEL_E, -1.6,
            Field.SIG_N, 0.7),
        "last", Map.of(
            "id", "AC42",
            "coords", Location.create(-162.78, 54.47),
            Field.SITETYPE, "cGPS",
            Field.VEL_E, -5.8,
            Field.SIG_N, 0.5)));

    // HI_2020
    expectedMap.put(GpsDataSet.HI_2020, Map.of(
        "count", 115,
        "first", Map.of(
            "id", "0",
            "coords", Location.create(-155.26610, 19.37911),
            Field.VEL_E, 22.8000,
            Field.VEL_V, -22.7000,
            Field.SIG_N, 0.0800),
        "last", Map.of(
            "id", "114",
            "coords", Location.create(-157.92083, 21.31299),
            Field.VEL_E, 0.2000,
            Field.VEL_V, -0.3000,
            Field.SIG_N, 0.0200)));

    // CONUS_2014_WUS
    expectedMap.put(GpsDataSet.CONUS_2014_WUS, Map.of(
        "count", 3168,
        "first", Map.of(
            "id", "0",
            "coords", Location.create(-118.14610, 34.75410),
            Field.VEL_E, -13.5440,
            Field.SIG_N, 0.6900,
            Field.CORRELATION, 0.1700),
        "last", Map.of(
            "id", "3167",
            "coords", Location.create(-111.02040, 39.29330),
            Field.VEL_E, -0.3720,
            Field.SIG_N, 0.3000,
            Field.CORRELATION, 0.0310)));

    // CONUS_2014_UCERF
    expectedMap.put(GpsDataSet.CONUS_2014_UCERF, Map.of(
        "count", 1867,
        "first", Map.of(
            "id", "0",
            "coords", Location.create(-118.14600, 34.75400),
            Field.VEL_E, -12.6980,
            Field.SIG_N, 0.8000,
            Field.CORRELATION, 0.0010),
        "last", Map.of(
            "id", "1866",
            "coords", Location.create(-114.00000, 33.00000),
            Field.VEL_E, 0.0000,
            Field.SIG_N, 1.0000,
            Field.CORRELATION, 0.0000)));

    // CONUS_2023
    expectedMap.put(GpsDataSet.CONUS_2023, Map.of(
        "count", 4979,
        "first", Map.of(
            "id", "0",
            "coords", Location.create(-117.09320, 34.11640),
            Field.VEL_E, -15.1890,
            Field.SIG_N, 0.4960,
            Field.CORRELATION, 0.0010),
        "last", Map.of(
            "id", "4978",
            "coords", Location.create(-104.21150, 36.49130),
            Field.VEL_E, -0.7700,
            Field.SIG_N, 0.1700,
            Field.CORRELATION, 0.0350)));

    expected = Collections.unmodifiableMap(expectedMap);
  }

  /*
   * TODO: Is there a better way? This test current tests that the method
   * functions and validates a small subset of the JSON results, better would be
   * to use test resource files and run comprehensive tests against those,
   * possibly with equivalent JSON files for comparison to the test CSV files.
   */
  @SuppressWarnings("unchecked")
  @ParameterizedTest
  @EnumSource(value = GpsDataSet.class,
      names = { "AK_2020", "HI_2020", "CONUS_2014_WUS", "CONUS_2014_UCERF", "CONUS_2023" })
  final void getJsonTest(GpsDataSet gpsDataSet) {
    String json = dataSets.getJsonString(gpsDataSet);
    List<Feature> features = GeoJson.from(json).toFeatureCollection().features();
    int n = (int) expected.get(gpsDataSet).get("count");
    assertEquals(n, features.size());
    if (expected.get(gpsDataSet).containsKey("first")) {
      checkFeature(
          (Map<String, Object>) expected.get(gpsDataSet).get("first"),
          features.get(0));
    }
    if (expected.get(gpsDataSet).containsKey("last")) {
      checkFeature(
          (Map<String, Object>) expected.get(gpsDataSet).get("last"),
          features.get(n - 1));
    }
  }

  private static void checkFeature(Map<String, Object> expected, Feature f) {
    assertEquals(expected.get("id"), f.idAsString().orElseThrow());
    assertEquals(expected.get("id"),
        f.properties().get(Field.STATIONID, String.class).orElseThrow());

    assertEquals(expected.get("coords"), f.asPoint());

    for (String key : propertiesToCheck) {
      if (expected.containsKey(key)) {
        assertEquals(expected.get(key), f.properties().get(key, String.class).orElseThrow());
      } else {
        assertEquals(Optional.empty(), f.properties().get(key, String.class));
      }
    }

    for (String key : doublesToCheck) {
      if (expected.containsKey(key)) {
        assertEquals(expected.get(key), f.properties().get(key, Double.class).orElseThrow());
      } else {
        assertEquals(Optional.empty(), f.properties().get(key, Double.class));
      }
    }
  }

  @ParameterizedTest
  @EnumSource(value = GpsDataSet.class)
  final void getCsvTest(GpsDataSet gpsDataSet) throws IOException {

    String actualCsv = Resources.toString(
        Resources.getResource(GpsDataSet.class, Utils.GPS_RESOURCES + gpsDataSet.filename),
        StandardCharsets.UTF_8);

    assertEquals(actualCsv, dataSets.getCsv(gpsDataSet));
  }

}
