package gov.usgs.earthquake.nshmp.www.gulf;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertTrue;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtensionContext;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.ArgumentsProvider;

import com.google.common.io.Resources;

import gov.usgs.earthquake.nshmp.Text;
import gov.usgs.earthquake.nshmp.Text.Delimiter;
import gov.usgs.earthquake.nshmp.data.DelimitedData;
import gov.usgs.earthquake.nshmp.geo.Location;
import gov.usgs.earthquake.nshmp.geo.json.Feature;
import gov.usgs.earthquake.nshmp.www.Utils;
import gov.usgs.earthquake.nshmp.www.gulf.GulfData.GulfDataResponse;

public class GulfDataTests {

  @Test
  final void gulfDataTest() throws IOException {
    GulfData gulfData = GulfData.readGulfData();

    List<String> lines = readLinesToList();
    DelimitedData data = DelimitedData.comma(lines);

    data.records()
        .forEach(record -> {
          double lon = record.getDouble("longitude");
          double lat = record.getDouble("latitude");
          double expectedDepth = record.getDouble("depth");
          Location loc = Location.create(lon, lat);
          double actualDepth = gulfData.getGulfData(loc);
          assertEquals(expectedDepth, actualDepth);

          GulfDataResponse actual = gulfData.getGulfDataResponse(loc);
          assertEquals(expectedDepth, actual.sedimentDepth);
          assertEquals(lon, actual.longitude);
          assertEquals(lat, actual.latitude);

          Feature actualFeature = gulfData.getGulfDataFeature(loc);
          assertEquals(expectedDepth,
              actualFeature.properties().getDouble("sedimentDepth").getAsDouble());
          assertEquals(lon, actualFeature.asPoint().longitude);
          assertEquals(lat, actualFeature.asPoint().latitude);
        });
  }

  private static List<String> readLinesToList() {
    List<String> lines = null;
    try {
      lines = Resources.readLines(
          Resources.getResource(GulfData.class, Utils.GULF_RESOURCES +
              GulfData.GULF_DATA_FILENAME),
          StandardCharsets.UTF_8);
    } catch (IOException e) {
      e.printStackTrace();
    }
    return lines;
  }

  public static class GulfParameterizedTests implements ArgumentsProvider {
    @Override
    public Stream<? extends Arguments> provideArguments(ExtensionContext context) throws Exception {

      List<String> lines = readLinesToList();

      List<Object[]> list = new ArrayList<>();

      list = lines.stream().skip(3).map(line -> {
        Iterator<String> it = Text.split(line, Delimiter.COMMA).iterator();
        return new Object[] {
            Double.valueOf(it.next()),
            Double.valueOf(it.next()),
            Double.valueOf(it.next())
        };

      }).collect(Collectors.toList());

      return list.stream().map(Arguments::of);
    }

  }

  @Test
  final void roundToGridTest() {
    double tol = 1e-6;
    double delta = 0.05;
    double x = delta / 5;
    // Rounding the halfway value appears to be handled differently for positive
    // and negative numbers, so a step of delta/2 is not tested here now
    // i.e. test passes for longitude test values including a step of delta / 2,
    // but test fails with latitude values with this step
    // Positive numbers appear to be rounded using the round-to-even strategy
    // for mid-points, while negative numbers seem to use the round towards zero
    // strategy for mid-points
    List<Double> steps =
        Arrays.asList(0.00, x, 2 * x, delta * 0.49, delta * 0.51, 3 * x, 4 * x,
            delta * 0.99); // add `delta / 2` to test mid-point rounding

    List<Double> values = Arrays.asList(-101.50, -101.45, -101.40, -101.35, -101.30, -101.25,
        -101.20, -101.15, -101.10, -101.05, -101.00, -100.95, -100.90);
    testRoundToGrid(values, steps, delta, tol);

    values = Arrays.asList(32.45, 32.50, 32.55, 32.60, 32.65, 32.70, 32.75, 32.80, 32.85, 32.90,
        32.95, 33.00, 33.05, 33.10, 33.15, 33.20, 33.25);
    testRoundToGrid(values, steps, delta, tol);

  }

  private void testRoundToGrid(List<Double> minimums, List<Double> steps, double delta,
      double tol) {
    for (double min : minimums) {

      double max = min + delta;

      for (double step : steps) {
        double value = min + step;
        double rounded = GulfData.roundToGrid(value, delta);
        if (Math.abs(step - delta / 2) < tol) {
          System.out.printf("Min, max, target, rounded: %.3f, %.3f, %.3f, %.3f\n",
              min, max, value, rounded);
        }
        // assume rounding method - midpoint rounds up
        if (step < delta / 2) {
          // target value is less than 1/2 delta above the minimum, rounds down
          assertEquals(min, rounded, tol);
          assertTrue(rounded - min < delta / 2);
          assertTrue(max - rounded > delta / 2);
        } else {
          // target value is greater than or equal to 1/2 delta above minimum,
          // rounds up
          assertEquals(max, rounded, tol);
          assertTrue(rounded - min >= delta / 2);
          assertTrue(max - rounded <= delta / 2);
        }
      }
    }
  }

  /*
   * rounding results
   */
  public static void main(String[] args) {
    double start = -3.1;
    double delta = 0.05;
    for (int i = 0; i < 110; i++) {
      double value = start + delta * i + delta / 2;
      // rounding to 3 places changes results
      // value = Maths.round(value, 3);
      double rounded = GulfData.roundToGrid(value, delta);
      System.out.printf(
          "%12s %.4f %-12s\n",
          (rounded < value) ? String.format("%.2f <---", rounded) : " ",
          value,
          (rounded < value) ? " " : String.format("---> %.2f", rounded));
    }
  }

}
